#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------
"""
Tests the pysat parameters storage area
"""

import copy
from importlib import reload
import os
import pytest
import shutil
import tempfile

import pysat  # required for reimporting pysat
from pysat._params import Parameters  # required for eval statements
from pysat.tests.travisci_test_class import TravisCICleanSetup


class TestBasics():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # Store current pysat directory
        self.stored_params = copy.deepcopy(pysat.params)

        # Set up default values
        pysat.params.restore_defaults()

        # Get a temporary directory
        self.tempdir = tempfile.TemporaryDirectory()
        self.wd = os.getcwd()

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        pysat.params = copy.deepcopy(self.stored_params)
        pysat.params.store()
        reload(pysat)
        self.tempdir.cleanup()
        os.chdir(self.wd)

    @pytest.mark.parametrize("paths, check",
                             [('.', ['.']),
                              (os.path.join('.', 'hi'),
                               [os.path.join('.', 'hi')]),
                              (os.path.join('.', 'hi', ''),
                               [os.path.join('.', 'hi')]),
                              (os.path.join('.', ''), ['.']),
                              (['.', '.'], None)])
    def test_set_data_dirs(self, paths, check):
        """Update pysat directory via params"""
        if check is None:
            check = paths

        # Switch working directory to temp directory
        os.chdir(self.tempdir.name)

        # Assign path
        pysat.params['data_dirs'] = paths
        assert pysat.params['data_dirs'] == check

        # Check if next load of pysat remembers the change
        reload(pysat)
        assert pysat.params['data_dirs'] == check

    @pytest.mark.parametrize("path",
                             ['no_path',
                              'not_a_directory'])
    def test_set_data_dir_bad_directory(self, path):
        """Ensure you can't set data_dirs to a bad path"""
        with pytest.raises(ValueError) as excinfo:
            pysat.params['data_dirs'] = path
        assert str(excinfo.value).find("Invalid path") >= 0
        return

    def test_repr(self):
        """Test __repr__ method"""
        out = pysat.params.__repr__()
        assert out.find('Parameters(path=') >= 0

    def test_str(self):
        """Ensure str method works"""

        # Include a user parameter
        pysat.params['pysat_user_test_str'] = 'We are here.'

        out = str(pysat.params)
        # Confirm start of str
        assert out.find('pysat Parameters object') >= 0

        # Confirm that default, non-default, and user values present
        assert out.find('pysat settings') > 0
        assert out.find('Standard parameters:') > 0

        assert out.find('settings (non-default)') > 0
        assert out.find('Standard parameters (no defaults):') > 0

        assert out.find('user values') > 0
        assert out.find('User parameters:') > 0

    def test_restore_defaults(self):
        """Test restore_defaults works as intended"""

        # Get default value, as per setup
        default_val = pysat.params['update_files']

        # Change value to non-default
        pysat.params['update_files'] = not default_val

        # Restore defaults
        pysat.params.restore_defaults()

        # Ensure new value is the default
        assert pysat.params['update_files'] == default_val

        # Make sure that non-default values left as is
        assert pysat.params['data_dirs'] != []

    def test_update_standard_value(self):
        """Modify a pre-existing standard parameter value and ensure stored"""

        # Get default value, as per setup
        default_val = pysat.params['update_files']

        # Change value to non-default
        pysat.params['update_files'] = not pysat.params['update_files']

        # Ensure it is in memory
        assert pysat.params['update_files'] is not default_val

        # Get a new parameters instance and verify information is retained.
        # Using eval to ensure all settings with current pysat.params retained.
        new_params = eval(pysat.params.__repr__())
        assert new_params['update_files'] == pysat.params['update_files']

    def test_no_update_user_modules(self):
        """Ensure user_modules not modifiable via params"""

        # Attempt to change value
        with pytest.raises(ValueError) as err:
            pysat.params['user_modules'] = {}
        assert str(err).find('The pysat.utils.registry ') >= 0

    def test_add_user_parameter(self):
        """Add custom parameter and ensure present"""

        pysat.params['hi_there'] = 'hello there!'
        assert pysat.params['hi_there'] == 'hello there!'

        # Get a new parameters instance and verify information is retained
        # Using eval to ensure all settings with current pysat.params retained.
        new_params = eval(pysat.params.__repr__())
        assert new_params['hi_there'] == pysat.params['hi_there']

    def test_clear_and_restart(self):
        """Verify clear_and_restart method impacts all values"""

        pysat.params.clear_and_restart()

        # Check default value
        assert pysat.params['user_modules'] == {}

        # Check value without working default
        assert pysat.params['data_dirs'] == []

        return

    def test_bad_path_instantiation(self):
        """Ensure you can't use bad path when loading Parameters"""
        with pytest.raises(OSError) as excinfo:
            Parameters(path='./made_up_name')
        assert str(excinfo.value).find("Supplied path does not exist") >= 0


class TestCIonly(TravisCICleanSetup):
    """Tests where we mess with local settings.
    These only run in CI environments such as Travis and Appveyor to avoid
    breaking an end user's setup
    """

    # Set setup/teardown to the class defaults
    setup = TravisCICleanSetup.setup
    teardown = TravisCICleanSetup.teardown

    def test_settings_file_must_be_present(self, capsys):
        """Ensure pysat_settings.json must be present"""

        reload(pysat)

        captured = capsys.readouterr()
        # Ensure pysat is running in 'first-time' mode
        assert captured.out.find("Hi there!") >= 0

        # Remove pysat settings file
        shutil.move(os.path.join(self.root, 'pysat_settings.json'),
                    os.path.join(self.root, 'pysat_settings_moved.json'))

        # Ensure we can't create a parameters file without valid .json
        with pytest.raises(OSError) as err:
            Parameters()
        assert str(err).find('pysat is unable to locate a user settings') >= 0

        shutil.move(os.path.join(self.root, 'pysat_settings_moved.json'),
                    os.path.join(self.root, 'pysat_settings.json'))

    def test_settings_file_cwd(self, capsys):
        """Test Parameters works when settings file in current working dir"""

        reload(pysat)

        captured = capsys.readouterr()
        # Ensure pysat is running in 'first-time' mode
        assert captured.out.find("Hi there!") >= 0

        # Move pysat settings file to cwd
        shutil.move(os.path.join(self.root, 'pysat_settings.json'),
                    os.path.join('./', 'pysat_settings.json'))

        # Try loading by supplying a specific path
        test_params = Parameters(path='./')

        # Supplying no path should yield the same result
        test_params2 = Parameters()

        # Confirm data is the same for both
        assert test_params.data == test_params2.data

        # Confirm path is the same for both
        assert test_params.file_path == test_params2.file_path

        # Ensure we didn't load a file in .pysat
        assert not os.path.isfile(os.path.join(self.root,
                                               'pysat_settings.json'))

        # Move pysat settings file back to original
        shutil.move(os.path.join('./', 'pysat_settings.json'),
                    os.path.join(self.root, 'pysat_settings.json'))



# -*- coding: utf-8 -*-
"""Supports the Langmuir Probe (LANG) instrument on Dynamics Explorer 2 (DE2).

From CDAWeb:

The Langmuir Probe Instrument (LANG) was a cylindrical electrostatic probe that
obtained measurements of electron temperature, Te, and electron or ion
concentration, Ne or Ni, respectively, and spacecraft potential.  Data from
this investigation were used to provide temperature and density measurements
along magnetic field lines related to thermal energy and particle flows within
the magnetosphere-ionosphere system, to provide thermal plasma conditions for
wave-particle interactions, and to measure large-scale and fine-structure
ionospheric effects of energy deposition in the ionosphere.  The Langmuir Probe
instrument was identical to that used on the AE satellites and the Pioneer
Venus Orbiter. Two independent sensors were connected to individual adaptive
sweep voltage circuits which continuously tracked the changing electron
temperature and spacecraft potential, while autoranging electrometers adjusted
their gain in response to the changing plasma density. The control signals used
to achieve this automatic tracking provided a continuous monitor of the
ionospheric parameters without telemetering each volt-ampere (V-I) curve.
Furthermore, internal data storage circuits permitted high resolution, high
data rate sampling of selected V-I curves for transmission to ground to verify
or correct the inflight processed data. Time resolution was 0.5 seconds.


References
----------
J. P. Krehbiel, L. H. Brace, R. F. Theis, W. H. Pinkus, and R. B. Kaplan,
The Dynamics Explorer 2 Langmuir Probe (LANG), Space Sci. Instrum., v. 5, n. 4,
p. 493, 1981.

Properties
----------
platform
    'de2'
name
    'lang'
sat_id
    None Supported
tag
    None Supported

Authors
-------
J. Klenzing

"""

from __future__ import print_function
from __future__ import absolute_import

import functools

import pysat
from pysat.instruments.methods import nasa_cdaweb as cdw
from pysat.instruments.methods import general as mm_gen

platform = 'de2'
name = 'lang'

tags = {'': '500 ms cadence Langmuir Probe data'}
sat_ids = {'': ['']}
_test_dates = {'': {'': pysat.datetime(1983, 1, 1)}}

fname = 'de2_plasma500ms_lang_{year:04d}{month:02d}{day:02d}_v01.cdf'
supported_tags = {'': {'': fname}}

# use the CDAWeb methods list files routine
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)

# use the default CDAWeb method
load = cdw.load

# support download routine
basic_tag = {'dir': '/pub/data/de/de2/plasma_lang/plasma500ms_lang_cdaweb',
             'remote_fname': '{year:4d}/' + fname,
             'local_fname': fname}
supported_tags = {'': {'': basic_tag}}
download = functools.partial(cdw.download, supported_tags)

# support listing files currently on CDAWeb
list_remote_files = functools.partial(cdw.list_remote_files,
                                      supported_tags=supported_tags)


def clean(inst):
    """Routine to return PLATFORM/NAME data cleaned to the specified level

    Cleaning level is specified in inst.clean_level and pysat
    will accept user input for several strings. The clean_level is
    specified at instantiation of the Instrument object.

    'clean' All parameters should be good, suitable for statistical and
            case studies
    'dusty' All paramers should generally be good though same may
            not be great
    'dirty' There are data areas that have issues, data should be used
            with caution
    'none'  No cleaning applied, routine not called in this case.


    Parameters
    -----------
    inst : pysat.Instrument
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    """

    return



# -*- coding: utf-8 -*-
"""Supports the Michelson Interferometer for Global High-resolution
Thermospheric Imaging (MIGHTI) instrument onboard the Ionospheric
CONnection Explorer (ICON) satellite.  Accesses local data in
netCDF format.

Properties
----------
platform
    'icon'
name
    'mighti'
tag
    Supports 'los_wind_green', 'los_wind_red', 'vector_wind_green',
    'vector_wind_red', 'temperature'.  Note that not every data product
    available for every sat_id
sat_id
    '', 'a', or 'b'


Warnings
--------
- The cleaning parameters for the instrument are still under development.
- Only supports level-2 data.


Example
-------
::

    import pysat
    mighti = pysat.Instrument('icon', 'mighti', 'vector_wind_green',
                              clean_level='clean')
    mighti.download(dt.datetime(2020, 1, 30), dt.datetime(2020, 1, 31))
    mighti.load(2020, 2)

By default, pysat removes the ICON level tags from variable names, ie,
ICON_L27_Ion_Density becomes Ion_Density.  To retain the original names, use
::

    mighti = pysat.Instrument(platform='icon', name='mighti',
                              tag='vector_wind_green', clean_level='clean',
                              keep_original_names=True)


Authors
---------
Originated from EUV support.
Jeff Klenzing, Mar 17, 2018, Goddard Space Flight Center
Russell Stoneback, Mar 23, 2018, University of Texas at Dallas
Conversion to MIGHTI, Oct 8th, 2028, University of Texas at Dallas

"""

from __future__ import print_function
from __future__ import absolute_import

import datetime as dt
import functools
import logging

import pysat
from pysat.instruments.methods import general as mm_gen
from pysat.instruments.methods import icon as mm_icon


logger = logging.getLogger(__name__)

platform = 'icon'
name = 'mighti'
tags = {'los_wind_green': 'Line of sight wind data -- Green Line',
        'los_wind_red': 'Line of sight wind data -- Red Line',
        'vector_wind_green': 'Vector wind data -- Green Line',
        'vector_wind_red': 'Vector wind data -- Red Line',
        'temperature': 'Neutral temperature data'}
sat_ids = {'': ['vector_wind_green', 'vector_wind_red'],
           'a': ['los_wind_green', 'los_wind_red', 'temperature'],
           'b': ['los_wind_green', 'los_wind_red', 'temperature']}
_test_dates = {jj: {kk: dt.datetime(2020, 1, 2) for kk in sat_ids[jj]}
               for jj in sat_ids.keys()}
_test_download_travis = {jj: {kk: False for kk in sat_ids[jj]}
                         for jj in sat_ids.keys()}
pandas_format = False

datestr = '{year:04d}-{month:02d}-{day:02d}_v{version:02d}r{revision:03d}'
fname1 = 'ICON_L2-1_MIGHTI-{id:s}_LOS-Wind-{color:s}_{date:s}.NC'
fname2 = 'ICON_L2-2_MIGHTI_Vector-Wind-{color:s}_{date:s}.NC'
fname3 = 'ICON_L2-3_MIGHTI-{id:s}_Temperature_{date:s}.NC'
supported_tags = {'': {'vector_wind_green': fname2.format(color='Green',
                                                          date=datestr),
                       'vector_wind_red': fname2.format(color='Red',
                                                        date=datestr)},
                  'a': {'los_wind_green': fname1.format(id='A', color='Green',
                                                        date=datestr),
                        'los_wind_red': fname1.format(id='A', color='Red',
                                                      date=datestr),
                        'temperature': fname3.format(id='A', date=datestr)},
                  'b': {'los_wind_green': fname1.format(id='B', color='Green',
                                                        date=datestr),
                        'los_wind_red': fname1.format(id='B', color='Red',
                                                      date=datestr),
                        'temperature': fname3.format(id='B', date=datestr)}}

# use the CDAWeb methods list files routine
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)

# support download routine
dirstr = '/pub/LEVEL.2/MIGHTI{id:s}'
dirdatestr = '{year:4d}/{doy:03d}/'
ids = {'': '',
       'a': '-A',
       'b': '-B'}
products = {'vector_wind_green': 'Vector-Winds/',
            'vector_wind_red': 'Vector-Winds/',
            'los_wind_green': 'LOS-Winds/',
            'los_wind_red': 'LOS-Winds/',
            'temperature': 'Temperature/'}
datestr = '{year:04d}-{month:02d}-{day:02d}'

download_tags = {}
for skey in supported_tags.keys():
    download_tags[skey] = {}
    for tkey in supported_tags[skey].keys():
        fname = supported_tags[skey][tkey]

        download_tags[skey][tkey] = {'dir': dirstr.format(id=ids[skey]),
                                     'remote_fname': ''.join((products[tkey],
                                                              fname))}

download = functools.partial(mm_icon.ssl_download, supported_tags=download_tags)

# support listing files on SSL
list_remote_files = functools.partial(mm_icon.list_remote_files,
                                      supported_tags=download_tags)


def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.

    Parameters
    -----------
    inst : (pysat.Instrument)
        Instrument class object

    """

    logger.info(mm_icon.ackn_str)
    self.meta.acknowledgements = mm_icon.ackn_str
    self.meta.references = ''.join((mm_icon.refs['mission'],
                                    mm_icon.refs['mighti']))

    pass


def default(inst):
    """Default routine to be applied when loading data.  Adjusts epoch timestamps
    to datetimes and removes variable preambles.

    """

    mm_gen.convert_timestamp_to_datetime(inst, sec_mult=1.0e-3)
    if not inst.kwargs['keep_original_names']:
        remove_preamble(inst)


def remove_preamble(inst):
    """Removes preambles in variable names"""

    target = {'los_wind_green': 'ICON_L21_',
              'los_wind_red': 'ICON_L21_',
              'vector_wind_green': 'ICON_L22_',
              'vector_wind_red': 'ICON_L22_',
              'temperature': ['ICON_L1_MIGHTI_{}_'.format(inst.sat_id.upper()),
                              'ICON_L23_MIGHTI_{}_'.format(inst.sat_id.upper()),
                              'ICON_L23_']}
    mm_gen.remove_leading_text(inst, target=target[inst.tag])

    return


def load(fnames, tag=None, sat_id=None, keep_original_names=False):
    """Loads ICON FUV data using pysat into pandas.

    This routine is called as needed by pysat. It is not intended
    for direct user interaction.

    Parameters
    ----------
    fnames : array-like
        iterable of filename strings, full path, to data files to be loaded.
        This input is nominally provided by pysat itself.
    tag : string
        tag name used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself.
    sat_id : string
        Satellite ID used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself.
    keep_original_names : boolean
        if True then the names as given in the netCDF ICON file
        will be used as is. If False, a preamble is removed.

    Returns
    -------
    data, metadata
        Data and Metadata are formatted for pysat. Data is a pandas
        DataFrame while metadata is a pysat.Meta instance.

    Note
    ----
    Any additional keyword arguments passed to pysat.Instrument
    upon instantiation are passed along to this routine.

    Examples
    --------
    ::
        inst = pysat.Instrument('icon', 'fuv')
        inst.load(2020, 1)

    """

    data, mdata = pysat.utils.load_netcdf4(fnames, epoch_name='Epoch',
                                           units_label='Units',
                                           name_label='Long_Name',
                                           notes_label='Var_Notes',
                                           desc_label='CatDesc',
                                           plot_label='FieldNam',
                                           axis_label='LablAxis',
                                           scale_label='ScaleTyp',
                                           min_label='ValidMin',
                                           max_label='ValidMax',
                                           fill_label='FillVal',
                                           pandas_format=pandas_format)
    # xarray can't merge if variable and dim names are the same
    if 'Altitude' in data.dims:
        data = data.rename({'Altitude': 'Alt'})
    return data, mdata


def clean(inst):
    """Provides data cleaning based upon clean_level.

    clean_level is set upon Instrument instantiation to
    one of the following:

    'Clean'
    'Dusty'
    'Dirty'
    'None'

    Routine is called by pysat, and not by the end user directly.

    Parameters
    -----------
    inst : (pysat.Instrument)
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    Note
    ----
        Supports 'clean', 'dusty', 'dirty', 'none'

    """

    if inst.tag.find('los') >= 0:
        # dealing with LOS winds
        wind_flag = 'Wind_Quality'
        ver_flag = 'VER_Quality'
        wind_vars = ['Line_of_Sight_Wind', 'Line_of_Sight_Wind_Error']
        ver_vars = ['Fringe_Amplitude', 'Fringe_Amplitude_Error',
                    'Relative_VER', 'Relative_VER_Error']
        if wind_flag not in inst.variables:
            wind_flag = '_'.join(('ICON_L21', wind_flag))
            ver_flag = '_'.join(('ICON_L21', ver_flag))
            wind_vars = ['ICON_L21_' + var for var in wind_vars]
            ver_vars = ['ICON_L21_' + var for var in ver_vars]
        min_val = {'clean': 1.0,
                   'dusty': 0.5}
        if inst.clean_level in ['clean', 'dusty']:
            # find location with any of the flags set
            for var in wind_vars:
                inst[var] = inst[var].where(inst[wind_flag]
                                            >= min_val[inst.clean_level])
            for var in ver_vars:
                inst[var] = inst[var].where(inst[ver_flag]
                                            >= min_val[inst.clean_level])
        else:
            # dirty and worse lets everything through
            pass

    elif inst.tag.find('vector') >= 0:
        # vector winds area
        wind_flag = 'Wind_Quality'
        ver_flag = 'VER_Quality'
        wind_vars = ['Zonal_Wind', 'Zonal_Wind_Error',
                     'Meridional_Wind', 'Meridional_Wind_Error']
        ver_vars = ['Fringe_Amplitude', 'Fringe_Amplitude_Error',
                    'Relative_VER', 'Relative_VER_Error']
        if wind_flag not in inst.variables:
            wind_flag = '_'.join(('ICON_L22', wind_flag))
            ver_flag = '_'.join(('ICON_L22', ver_flag))
            wind_vars = ['ICON_L22_' + var for var in wind_vars]
            ver_vars = ['ICON_L22_' + var for var in ver_vars]
        min_val = {'clean': 1.0,
                   'dusty': 0.5}
        if inst.clean_level in ['clean', 'dusty']:
            # find location with any of the flags set
            for var in wind_vars:
                inst[var] = inst[var].where(inst[wind_flag]
                                            >= min_val[inst.clean_level])
            for var in ver_vars:
                inst[var] = inst[var].where(inst[ver_flag]
                                            >= min_val[inst.clean_level])
        else:
            # dirty lets everything through
            pass

    elif inst.tag.find('temp') >= 0:
        # neutral temperatures
        var = 'Temperature'
        saa_flag = 'Quality_Flag_South_Atlantic_Anomaly'
        cal_flag = 'Quality_Flag_Bad_Calibration'
        if saa_flag not in inst.variables:
            saa_flag = '_'.join(('ICON_L1_MIGHTI', inst.sat_id.upper(),
                                 saa_flag))
            cal_flag = '_'.join(('ICON_L1_MIGHTI', inst.sat_id.upper(),
                                 cal_flag))
            var = '_'.join(('ICON_L23_MIGHTI', inst.sat_id.upper(), var))
        if inst.clean_level in ['clean', 'dusty']:
            # filter out areas with bad calibration data
            # as well as data marked in the SAA
            inst[var] = inst[var].where((inst[saa_flag] == 0)
                                        & (inst[cal_flag] == 0))
            # filter out negative temperatures
            inst[var] = inst[var].where(inst[var] > 0)
        else:
            # dirty and worse lets everything through
            pass

    return



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup
# To use a consistent encoding
import codecs
import os
import sys

here = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(here, 'description.txt'), encoding='utf-8') as f:
    long_description = f.read()
version_filename = os.path.join('pysat', 'version.txt')
with codecs.open(os.path.join(here, version_filename)) as version_file:
    version = version_file.read().strip()

# packages to be installed
# starting with packages common across all setups
install_requires = ['requests', 'beautifulsoup4', 'lxml', 'netCDF4']
# packages with Fortran code
fortran_install = ['pysatCDF', 'madrigalWeb', 'h5py', 'PyForecastTools']
# python version specific support libraries
if sys.version_info.major == 2:
    install_requires.extend(['xarray<0.12', 'pandas>=0.23, <0.25',
                             'numpy>=1.12, <1.17', 'scipy<1.3',
                             'matplotlib<3.0'])
else:
    # python 3+
    install_requires.extend(['xarray<0.15', 'pandas>=0.23, <0.25',
                             'numpy>=1.12', 'scipy', 'matplotlib'])

# flag, True if on readthedocs
on_rtd = os.environ.get('READTHEDOCS') == 'True'

# include Fortran for normal install
# read the docs doesn't do Fortran
if not on_rtd:
    # not on ReadTheDocs, add Fortran
    install_requires.extend(fortran_install)

setup(
    name='pysat',
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,

    description='Supports science data analysis across measurement platforms',
    long_description=long_description,
    # The project's main homepage.
    url='http://github.com/pysat/pysat',

    # Author details
    author='Russell Stoneback',
    author_email='rstoneba@utdallas.edu',


    package_data={'pysat': ['pysat/version*.txt']},
    include_package_data=True,

    # Choose your license
    license='BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Atmospheric Science',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # What does your project relate to?
    # keywords='sample setuptools development',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=['pysat', 'pysat.instruments', 'pysat.ssnl'],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=install_requires,
)



#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------
"""
tests the pysat utils area
"""

import contextlib
from io import StringIO
from importlib import reload
import numpy as np
import os
import portalocker
import pytest
import shutil
import tempfile

import pysat
from pysat.tests.registration_test_class import TestWithRegistration


# ----------------------------------
# test netCDF export file support
def prep_dir(inst=None):

    if inst is None:
        inst = pysat.Instrument(platform='pysat', name='testing')
    # create data directories
    try:
        os.makedirs(inst.files.data_path)
    except OSError:
        pass


def remove_files(inst):
    # remove any files
    temp_dir = inst.files.data_path
    for the_file in os.listdir(temp_dir):
        if (the_file == 'pysat_test_ncdf.nc'):
            file_path = os.path.join(temp_dir, the_file)
            if os.path.isfile(file_path):
                os.unlink(file_path)


class TestBasics():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # store current pysat directory
        self.data_path = pysat.params['data_dirs']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        pysat.params['data_dirs'] = self.data_path


class TestCIonly():
    """Tests where we mess with local settings.
    These only run in CI environments such as Travis and Appveyor to avoid
    breaking an end user's setup
    """

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.ci_env = (os.environ.get('TRAVIS') == 'true')
        if not self.ci_env:
            pytest.skip("Skipping local tests to avoid breaking user setup")

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.ci_env

    def test_initial_pysat_load(self, capsys):
        """Ensure initial load routines work"""

        # Move settings directory to simulate first load after install
        root = os.path.join(os.getenv('HOME'), '.pysat')
        new_root = os.path.join(os.getenv('HOME'), '.saved_pysat')
        shutil.move(root, new_root)

        reload(pysat)

        captured = capsys.readouterr()
        assert captured.out.find("Hi there!") >= 0

        # Make sure settings file created
        assert os.path.isfile(os.path.join(root, 'pysat_settings.json'))
        assert os.path.isdir(os.path.join(root, 'instruments'))
        assert os.path.isdir(os.path.join(root, 'instruments', 'archive'))

        # Move settings back
        shutil.rmtree(root)
        shutil.move(new_root, root)


class TestScaleUnits():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.deg_units = ["deg", "degree", "degrees", "rad", "radian",
                          "radians", "h", "hr", "hrs", "hours"]
        self.dist_units = ["m", "km", "cm"]
        self.vel_units = ["m/s", "cm/s", "km/s", 'm s$^{-1}$', 'cm s$^{-1}$',
                          'km s$^{-1}$', 'm s-1', 'cm s-1', 'km s-1']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.deg_units, self.dist_units, self.vel_units

    def test_scale_units_same(self):
        """ Test scale_units when both units are the same """

        scale = pysat.utils.scale_units("happy", "happy")

        assert scale == 1.0

    def test_scale_units_angles(self):
        """Test scale_units for angles """

        for out_unit in self.deg_units:
            scale = pysat.utils.scale_units(out_unit, "deg")

            if out_unit.find("deg") == 0:
                assert scale == 1.0
            elif out_unit.find("rad") == 0:
                assert scale == np.pi / 180.0
            else:
                assert scale == 1.0 / 15.0

    def test_scale_units_dist(self):
        """Test scale_units for distances """

        for out_unit in self.dist_units:
            scale = pysat.utils.scale_units(out_unit, "m")

            if out_unit == "m":
                assert scale == 1.0
            elif out_unit.find("km") == 0:
                assert scale == 0.001
            else:
                assert scale == 100.0

    def test_scale_units_vel(self):
        """Test scale_units for velocities """

        for out_unit in self.vel_units:
            scale = pysat.utils.scale_units(out_unit, "m/s")

            if out_unit.find("m") == 0:
                assert scale == 1.0
            elif out_unit.find("km") == 0:
                assert scale == 0.001
            else:
                assert scale == 100.0

    def test_scale_units_bad_output(self):
        """Test scale_units for unknown output unit"""

        with pytest.raises(ValueError) as verr:
            pysat.utils.scale_units('happy', 'm')
        assert str(verr).find('output unit') > 0

    def test_scale_units_bad_input(self):
        """Test scale_units for unknown input unit"""

        with pytest.raises(ValueError) as verr:
            pysat.utils.scale_units('m', 'happy')
        assert str(verr).find('input unit') > 0

    @pytest.mark.parametrize("unit1,unit2", [("m", "m/s"),
                                             ("m", "deg"),
                                             ("h", "km/s")])
    def test_scale_units_bad_match_pairs(self, unit1, unit2):
        """Test scale_units for mismatched input for all pairings"""

        with pytest.raises(ValueError):
            pysat.utils.scale_units(unit1, unit2)

    def test_scale_units_bad_match_message(self):
        """Test scale_units error message for mismatched input"""

        with pytest.raises(ValueError) as verr:
            pysat.utils.scale_units('m', 'm/s')
        assert str(verr).find('Cannot scale') >= 0
        assert str(verr).find('unknown units') < 0

    def test_scale_units_both_bad(self):
        """Test scale_units for bad input and output"""

        with pytest.raises(ValueError) as verr:
            pysat.utils.scale_units('happy', 'sad')
        assert str(verr).find('unknown units') > 0


class TestBasicNetCDF4():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # store current pysat directory
        self.data_path = pysat.params['data_dirs']

        # create temporary directory
        self.tempdir = tempfile.TemporaryDirectory()
        pysat.params['data_dirs'] = [self.tempdir.name]

        self.testInst = pysat.Instrument(platform='pysat', name='testing',
                                         num_samples=100, update_files=True)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']
        self.testInst.pandas_format = True

        # create testing directory
        prep_dir(self.testInst)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        pysat.params['data_dirs'] = self.data_path
        self.tempdir.cleanup()
        del self.testInst, self.stime

    def test_load_netcdf4_empty_filenames(self):
        with pytest.raises(ValueError):
            pysat.utils.load_netcdf4(fnames=None)

    @pytest.mark.parametrize('unlimited', [True, False])
    def test_basic_write_and_read_netcdf4_default_format(self, unlimited):
        """Test writing and loading netcdf4 file, with/out unlimited time dim
        """
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(date=self.stime)
        self.testInst.to_netcdf4(outfile, unlimited_time=unlimited)

        loaded_inst, meta = \
            pysat.utils.load_netcdf4(outfile,
                                     pandas_format=self.testInst.pandas_format)
        self.testInst.data = \
            self.testInst.data.reindex(sorted(self.testInst.data.columns),
                                       axis=1)
        loaded_inst = loaded_inst.reindex(sorted(loaded_inst.columns),
                                          axis=1)

        for key in self.testInst.data.columns:
            assert(np.all(self.testInst[key] == loaded_inst[key]))

    def test_basic_write_and_read_netcdf4_mixed_case_format(self):
        """ Test basic netCDF4 read/write with mixed case data variables
        """
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(date=self.stime)
        # modify data names in data
        original = sorted(self.testInst.data.columns)
        self.testInst.data = self.testInst.data.rename(str.upper,
                                                       axis='columns')
        self.testInst.to_netcdf4(outfile, preserve_meta_case=True)

        loaded_inst, meta = pysat.utils.load_netcdf4(outfile)
        self.testInst.data = self.testInst.data.reindex(
            sorted(self.testInst.data.columns), axis=1)
        loaded_inst = loaded_inst.reindex(sorted(loaded_inst.columns), axis=1)

        # check that names are lower case when written
        assert(np.all(original == loaded_inst.columns))

        for key in self.testInst.data.columns:
            assert(np.all(self.testInst[key] == loaded_inst[key.lower()]))

        # modify metadata names in data
        self.testInst.meta.data = self.testInst.meta.data.rename(str.upper,
                                                                 axis='index')
        # write file
        self.testInst.to_netcdf4(outfile, preserve_meta_case=True)
        # load file
        loaded_inst, meta = pysat.utils.load_netcdf4(outfile)

        # check that names are upper case when written
        assert np.all(sorted(self.testInst.data.columns)
                      == sorted(loaded_inst.columns))

    def test_write_netcdf4_duplicate_variable_names(self):
        """ Test netCDF4 writing with duplicate variable names
        """
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(date=self.stime)
        self.testInst['MLT'] = 1
        with pytest.raises(ValueError):
            self.testInst.to_netcdf4(outfile, preserve_meta_case=True)

    def test_write_and_read_netcdf4_default_format_w_compression(self):
        """Test success of writing and reading a compressed netCDF4 file
        """
        # Create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(date=self.stime)
        self.testInst.to_netcdf4(outfile, zlib=True)

        loaded_inst, meta = pysat.utils.load_netcdf4(outfile)
        self.testInst.data = self.testInst.data.reindex(
            sorted(self.testInst.data.columns), axis=1)
        loaded_inst = loaded_inst.reindex(sorted(loaded_inst.columns), axis=1)

        for key in self.testInst.data.columns:
            assert (np.all(self.testInst[key] == loaded_inst[key]))

    def test_write_and_read_netcdf4_default_format_w_weird_epoch_name(self):
        """ Test the netCDF4 write/read abilities with an odd epoch name
        """
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(date=self.stime)
        self.testInst.to_netcdf4(outfile, epoch_name='Santa')

        loaded_inst, meta = pysat.utils.load_netcdf4(outfile,
                                                     epoch_name='Santa')
        self.testInst.data = self.testInst.data.reindex(
            sorted(self.testInst.data.columns), axis=1)
        loaded_inst = loaded_inst.reindex(sorted(loaded_inst.columns), axis=1)

        for key in self.testInst.data.columns:
            assert (np.all(self.testInst[key] == loaded_inst[key]))

    def test_write_and_read_netcdf4_default_format_higher_order(self):
        # create a bunch of files by year and doy
        test_inst = pysat.Instrument('pysat', 'testing2d', update_files=True)
        prep_dir(test_inst)
        outfile = os.path.join(test_inst.files.data_path, 'pysat_test_ncdf.nc')
        test_inst.load(2009, 1)
        test_inst.to_netcdf4(outfile)
        loaded_inst, meta = pysat.utils.load_netcdf4(outfile)
        test_inst.data = test_inst.data.reindex(sorted(test_inst.data.columns),
                                                axis=1)
        loaded_inst = loaded_inst.reindex(sorted(loaded_inst.columns), axis=1)
        prep_dir(test_inst)

        # test Series of DataFrames
        test_list = []
        for frame1, frame2 in zip(test_inst.data['profiles'],
                                  loaded_inst['profiles']):
            test_list.append(np.all((frame1 == frame2).all()))
        loaded_inst.drop('profiles', inplace=True, axis=1)
        test_inst.data.drop('profiles', inplace=True, axis=1)

        # second series of frames
        for frame1, frame2 in zip(test_inst.data['alt_profiles'],
                                  loaded_inst['alt_profiles']):
            test_list.append(np.all((frame1 == frame2).all()))
        loaded_inst.drop('alt_profiles', inplace=True, axis=1)
        test_inst.data.drop('alt_profiles', inplace=True, axis=1)

        # check series of series
        for frame1, frame2 in zip(test_inst.data['series_profiles'],
                                  loaded_inst['series_profiles']):
            test_list.append(np.all((frame1 == frame2).all()))

        loaded_inst.drop('series_profiles', inplace=True, axis=1)
        test_inst.data.drop('series_profiles', inplace=True, axis=1)

        assert(np.all((test_inst.data == loaded_inst).all()))
        assert np.all(test_list)

    def test_write_and_read_netcdf4_default_format_higher_order_w_zlib(self):
        # create a bunch of files by year and doy
        test_inst = pysat.Instrument('pysat', 'testing2d', update_files=True)
        prep_dir(test_inst)
        outfile = os.path.join(test_inst.files.data_path, 'pysat_test_ncdf.nc')
        test_inst.load(2009, 1)
        test_inst.to_netcdf4(outfile, zlib=True)
        loaded_inst, meta = pysat.utils.load_netcdf4(outfile)
        test_inst.data = test_inst.data.reindex(sorted(test_inst.data.columns),
                                                axis=1)
        loaded_inst = loaded_inst.reindex(sorted(loaded_inst.columns), axis=1)
        prep_dir(test_inst)

        # test Series of DataFrames
        test_list = []
        for frame1, frame2 in zip(test_inst.data['profiles'],
                                  loaded_inst['profiles']):
            test_list.append(np.all((frame1 == frame2).all()))
        loaded_inst.drop('profiles', inplace=True, axis=1)
        test_inst.data.drop('profiles', inplace=True, axis=1)

        # second series of frames
        for frame1, frame2 in zip(test_inst.data['alt_profiles'],
                                  loaded_inst['alt_profiles']):
            test_list.append(np.all((frame1 == frame2).all()))
        loaded_inst.drop('alt_profiles', inplace=True, axis=1)
        test_inst.data.drop('alt_profiles', inplace=True, axis=1)

        # check series of series
        for frame1, frame2 in zip(test_inst.data['series_profiles'],
                                  loaded_inst['series_profiles']):
            test_list.append(np.all((frame1 == frame2).all()))
        loaded_inst.drop('series_profiles', inplace=True, axis=1)
        test_inst.data.drop('series_profiles', inplace=True, axis=1)

        assert (np.all((test_inst.data == loaded_inst).all()))
        assert np.all(test_list)

    def test_netcdf_prevent_attribute_override(self):
        """Test that attributes will not be overridden by default
        """
        self.testInst.load(date=self.stime)

        try:
            assert self.testInst.bespoke  # should raise
        except AttributeError:
            pass

        # instrument meta attributes immutable upon load
        assert not self.testInst.meta.mutable
        try:
            self.testInst.meta.bespoke = True
        except AttributeError:
            pass

    def test_netcdf_attribute_override(self):
        """Test that attributes in netcdf file may be overridden
        """
        self.testInst.load(date=self.stime)
        self.testInst.meta.mutable = True
        self.testInst.meta.bespoke = True

        self.testInst.meta.transfer_attributes_to_instrument(self.testInst)

        # Ensure custom meta attribute assigned to instrument
        assert self.testInst.bespoke

        fname = 'output.nc'
        outfile = os.path.join(self.testInst.files.data_path, fname)
        self.testInst.to_netcdf4(outfile)

        data, meta = pysat.utils.load_netcdf4(outfile)

        # Custom attribute correctly read from file
        assert meta.bespoke


class TestBasicNetCDF4xarray():
    """NOTE: combine with above class as part of #60"""

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # store current pysat directory
        self.data_path = pysat.params['data_dirs']

        # create temporary directory
        self.tempdir = tempfile.TemporaryDirectory()
        pysat.params['data_dirs'] = [self.tempdir.name]

        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing2d_xarray',
                                         update_files=True,
                                         num_samples=100)
        self.stime = pysat.instruments.pysat_testing2d_xarray._test_dates[
            '']['']

        # create testing directory
        prep_dir(self.testInst)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        # remove_files(self.testInst)
        pysat.params['data_dirs'] = self.data_path
        self.tempdir.cleanup()
        del self.testInst, self.stime

    def test_basic_write_and_read_netcdf4_default_format(self):
        """ Test basic netCDF4 writing and reading
        """
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(date=self.stime)
        self.testInst.data.attrs['new_attr'] = 1
        self.testInst.data.to_netcdf(outfile)

        loaded_inst, meta = pysat.utils.load_netcdf4(
            outfile, pandas_format=self.testInst.pandas_format)
        keys = self.testInst.data.data_vars.keys()

        for key in keys:
            assert(np.all(self.testInst[key] == loaded_inst[key]))
        assert meta.new_attr == 1

    def test_load_netcdf4_pandas_3d_error(self):
        """ Test load_netcdf4 error with a pandas 3D file
        """
        # create a bunch of files by year and doy
        prep_dir(self.testInst)
        outfile = os.path.join(self.testInst.files.data_path,
                               'pysat_test_ncdf.nc')
        self.testInst.load(date=self.stime)
        self.testInst.data.attrs['new_attr'] = 1
        self.testInst.data.to_netcdf(outfile)

        with pytest.raises(ValueError):
            loaded_inst, meta = pysat.utils.load_netcdf4(
                outfile, epoch_name='time', pandas_format=True)


class TestFmtCols():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # store current pysat directory
        self.in_str = np.arange(0, 40, 1).astype(str)
        self.in_kwargs = {"ncols": 5, "max_num": 40, "lpad": None}
        self.out_str = None
        self.filler_row = -1
        self.ncols = None
        self.nrows = None
        self.lpad = len(self.in_str[-1]) + 1

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.in_str, self.in_kwargs, self.out_str, self.filler_row
        del self.ncols, self.nrows, self.lpad

    def test_output(self):
        """ Test for the expected number of rows, columns, and fillers
        """
        if self.out_str is None and self.ncols is None and self.nrows is None:
            return

        # Test the number of rows
        out_rows = self.out_str.split('\n')[:-1]
        assert len(out_rows) == self.nrows

        # Test the number of columns
        for i, row in enumerate(out_rows):
            split_row = row.split()

            # Test for filler ellipses and standard row length
            if i == self.filler_row:
                assert '...' in split_row
                if i > 0:
                    assert len(split_row) == 1
                    assert len(row) == self.lpad * self.ncols
            else:
                assert len(row) == self.lpad * len(split_row)

                if i == len(out_rows) - 1:
                    assert len(split_row) <= self.ncols
                else:
                    assert len(split_row) == self.ncols

        return

    def test_neg_ncols(self):
        """ Test the output if the column number is negative
        """
        self.in_kwargs['ncols'] = -5
        self.out_str = pysat.utils._core.fmt_output_in_cols(self.in_str,
                                                            **self.in_kwargs)
        assert len(self.out_str) == 0

    @pytest.mark.parametrize("key,val,raise_type",
                             [("ncols", 0, ZeroDivisionError),
                              ("max_num", -10, ValueError)])
    def test_fmt_raises(self, key, val, raise_type):
        self.in_kwargs[key] = val
        with pytest.raises(raise_type):
            pysat.utils._core.fmt_output_in_cols(self.in_str, **self.in_kwargs)

    @pytest.mark.parametrize("ncol", [(3), (5), (10)])
    def test_ncols(self, ncol):
        """ Test the output for different number of columns
        """
        # Set the input
        self.in_kwargs['ncols'] = ncol

        # Set the comparison values
        self.ncols = ncol
        self.nrows = int(np.ceil(self.in_kwargs['max_num'] / ncol))

        # Get and test the output
        self.out_str = pysat.utils._core.fmt_output_in_cols(self.in_str,
                                                            **self.in_kwargs)
        self.test_output()

    @pytest.mark.parametrize("max_num,filler,nrow", [(0, 0, 1), (1, 0, 1),
                                                     (10, 1, 3), (50, -1, 8)])
    def test_max_num(self, max_num, filler, nrow):
        """ Test the output for the maximum number of values
        """
        # Set the input
        self.in_kwargs['max_num'] = max_num

        # Set the comparison values
        self.filler_row = filler
        self.ncols = self.in_kwargs['ncols']
        self.nrows = nrow

        # Get and test the output
        self.out_str = pysat.utils._core.fmt_output_in_cols(self.in_str,
                                                            **self.in_kwargs)
        self.test_output()

    @pytest.mark.parametrize("in_pad", [5, 30])
    def test_lpad(self, in_pad):
        """ Test the output for different number of columns
        """
        # Set the input
        self.in_kwargs['lpad'] = in_pad
        self.ncols = self.in_kwargs['ncols']
        self.nrows = int(np.ceil(self.in_kwargs['max_num'] / self.ncols))

        # Set the comparison values
        self.lpad = in_pad

        # Get and test the output
        self.out_str = pysat.utils._core.fmt_output_in_cols(self.in_str,
                                                            **self.in_kwargs)
        self.test_output()


class TestAvailableInst(TestWithRegistration):

    # Set setup/teardown to the class defaults
    setup = TestWithRegistration.setup
    teardown = TestWithRegistration.teardown

    @pytest.mark.parametrize("inst_loc", [None, pysat.instruments])
    @pytest.mark.parametrize("inst_flag, plat_flag",
                             [(None, None), (False, False), (True, True)])
    def test_display_available_instruments(self, inst_loc, inst_flag,
                                           plat_flag):
        """Test display_available_instruments options
        """
        # If using the pysat registry, make sure there is something registered
        if inst_loc is None:
            pysat.utils.registry.register(self.module_names)

        # Initialize the STDOUT stream
        new_stdout = StringIO()

        with contextlib.redirect_stdout(new_stdout):
            pysat.utils.display_available_instruments(
                inst_loc, show_inst_mod=inst_flag, show_platform_name=plat_flag)

        out = new_stdout.getvalue()
        assert out.find("Description") > 0

        if (inst_loc is None and plat_flag is None) or plat_flag:
            assert out.find("Platform") == 0
            assert out.find("Name") > 0

        if (inst_loc is not None and inst_flag is None) or inst_flag:
            assert out.find("Instrument_Module") >= 0

        if inst_loc is not None and inst_flag in [None, True]:
            assert out.find(inst_loc.__name__) > 0

        return

    def test_import_error_in_available_instruments(self):
        """ Test handling of import errors in available_instruments
        """

        idict = pysat.utils.available_instruments(os.path)

        for platform in idict.keys():
            for name in idict[platform].keys():
                assert 'ERROR' in idict[platform][name]['inst_ids_tags'].keys()
                assert 'ERROR' in idict[platform][name][
                    'inst_ids_tags']['ERROR']
        return


class TestNetworkLock():
    def setup(self):
        self.fname = 'temp_lock_file.txt'
        with open(self.fname, 'w') as fh:
            fh.write('spam and eggs')

    def teardown(self):
        os.remove(self.fname)

    def test_with_timeout(self):
        # Open the file 2 times
        with pytest.raises(portalocker.AlreadyLocked):
            with pysat.utils.NetworkLock(self.fname, timeout=0.1):
                with pysat.utils.NetworkLock(self.fname, mode='wb', timeout=0.1,
                                             fail_when_locked=True):
                    pass

    def test_without_timeout(self):
        # Open the file 2 times
        with pytest.raises(portalocker.LockException):
            with pysat.utils.NetworkLock(self.fname, timeout=None):
                with pysat.utils.NetworkLock(self.fname, timeout=None,
                                             mode='w'):
                    pass

    def test_without_fail(self):
        # Open the file 2 times
        with pytest.raises(portalocker.LockException):
            with pysat.utils.NetworkLock(self.fname, timeout=0.1):
                lock = pysat.utils.NetworkLock(self.fname, timeout=0.1)
                lock.acquire(check_interval=0.05, fail_when_locked=False)



# -*- coding: utf-8 -*-
# Test some of the basic _core functions
import numpy as np
import sys

from nose.tools import raises
import pandas as pds

import pysat
import pysat.instruments.pysat_testing
import pysat.instruments.pysat_testing_xarray
from pysat.utils.time import filter_datetime_input

if sys.version_info[0] >= 3:
    from importlib import reload as re_load
else:
    re_load = reload


# ------------------------------------------------------------------------------
#
# Test Instrument object basics
#
# ------------------------------------------------------------------------------

class TestBasics():
    def setup(self):
        re_load(pysat.instruments.pysat_testing)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat', name='testing',
                                         sat_id='10',
                                         clean_level='clean',
                                         update_files=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst

    # --------------------------------------------------------------------------
    #
    # Test basic loads, by date, filename, file id, as well as prev/next
    #
    # --------------------------------------------------------------------------
    def test_basic_instrument_load(self):
        """Test if the correct day is being loaded (checking object date and
        data)."""
        self.testInst.load(2009, 1)
        test_date = self.testInst.index[0]
        test_date = pysat.datetime(test_date.year, test_date.month,
                                   test_date.day)
        assert (test_date == pds.datetime(2009, 1, 1)) & \
            (test_date == self.testInst.date)

    @raises(Exception)
    def test_basic_instrument_bad_keyword(self):
        """Checks for error when instantiating with bad load_rtn keywords"""
        testInst = pysat.Instrument(platform='pysat', name='testing',
                                    sat_id='10',
                                    clean_level='clean',
                                    unsupported_keyword_yeah=True)

    @raises(Exception)
    def test_basic_instrument_load_yr_no_doy(self):
        self.testInst.load(2009)

    @raises(Exception)
    def test_basic_instrument_load_no_input(self):
        self.testInst.load()

    @raises(Exception)
    def test_basic_instrument_load_by_file_and_multifile(self):
        testInst = pysat.Instrument(platform='pysat', name='testing',
                                    sat_id='10',
                                    clean_level='clean',
                                    update_files=True,
                                    multi_file_day=True)
        testInst.load(fname=testInst.files[0])

    def test_basic_instrument_load_by_date(self):
        date = pysat.datetime(2009, 1, 1)
        self.testInst.load(date=date)
        test_date = self.testInst.index[0]
        test_date = pysat.datetime(test_date.year, test_date.month,
                                   test_date.day)
        assert (test_date == pds.datetime(2009, 1, 1))
        assert (test_date == self.testInst.date)

    def test_basic_instrument_load_by_date_with_extra_time(self):
        # put in a date that has more than year, month, day
        date = pysat.datetime(2009, 1, 1, 1, 1, 1)
        self.testInst.load(date=date)
        test_date = self.testInst.index[0]
        test_date = pysat.datetime(test_date.year, test_date.month,
                                   test_date.day)
        assert (test_date == pds.datetime(2009, 1, 1))
        assert (test_date == self.testInst.date)

    def test_basic_instrument_load_data(self):
        """Test if the correct day is being loaded (checking data down to the
        second)."""
        self.testInst.load(2009, 1)
        assert self.testInst.index[0] == pds.datetime(2009, 1, 1, 0, 0, 0)

    def test_basic_instrument_load_leap_year(self):
        """Test if the correct day is being loaded (Leap-Year)."""
        self.testInst.load(2008, 366)
        test_date = self.testInst.index[0]
        test_date = pysat.datetime(test_date.year, test_date.month,
                                   test_date.day)
        assert (test_date == pds.datetime(2008, 12, 31)) & \
            (test_date == self.testInst.date)

    def test_next_load_default(self):
        """Test if first day is loaded by default when first invoking .next."""
        self.testInst.next()
        test_date = self.testInst.index[0]
        test_date = pysat.datetime(test_date.year, test_date.month,
                                   test_date.day)
        assert test_date == pds.datetime(2008, 1, 1)

    def test_prev_load_default(self):
        """Test if last day is loaded by default when first invoking .prev."""
        self.testInst.prev()
        test_date = self.testInst.index[0]
        test_date = pysat.datetime(test_date.year, test_date.month,
                                   test_date.day)
        assert test_date == pds.datetime(2010, 12, 31)

    def test_basic_fid_instrument_load(self):
        """Test if first day is loaded by default when first invoking .next."""
        self.testInst.load(fid=0)
        test_date = self.testInst.index[0]
        test_date = pysat.datetime(test_date.year, test_date.month,
                                   test_date.day)
        assert (test_date == pds.datetime(2008, 1, 1)) & \
            (test_date == self.testInst.date)

    def test_next_fid_load_default(self):
        """Test next day is being loaded (checking object date)."""
        self.testInst.load(fid=0)
        self.testInst.next()
        test_date = self.testInst.index[0]
        test_date = pysat.datetime(test_date.year, test_date.month,
                                   test_date.day)
        assert (test_date == pds.datetime(2008, 1, 2)) & \
            (test_date == self.testInst.date)

    def test_prev_fid_load_default(self):
        """Test prev day is loaded when invoking .prev."""
        self.testInst.load(fid=3)
        self.testInst.prev()
        test_date = self.testInst.index[0]
        test_date = pysat.datetime(test_date.year, test_date.month,
                                   test_date.day)
        assert (test_date == pds.datetime(2008, 1, 3)) & \
            (test_date == self.testInst.date)

    def test_filename_load(self):
        """Test if file is loadable by filename, relative to
        top_data_dir/platform/name/tag"""
        self.testInst.load(fname='2010-12-31.nofile')
        assert self.testInst.index[0] == pds.datetime(2010, 12, 31)

    def test_next_filename_load_default(self):
        """Test next day is being loaded (checking object date)."""
        self.testInst.load(fname='2010-12-30.nofile')
        self.testInst.next()
        test_date = self.testInst.index[0]
        test_date = pysat.datetime(test_date.year, test_date.month,
                                   test_date.day)
        assert (test_date == pds.datetime(2010, 12, 31)) & \
            (test_date == self.testInst.date)

    def test_prev_filename_load_default(self):
        """Test prev day is loaded when invoking .prev."""
        self.testInst.load(fname='2009-01-04.nofile')
        # print(self.testInst.date)
        self.testInst.prev()
        test_date = self.testInst.index[0]
        test_date = pysat.datetime(test_date.year, test_date.month,
                                   test_date.day)
        assert (test_date == pds.datetime(2009, 1, 3))
        assert (test_date == self.testInst.date)

    # --------------------------------------------------------------------------
    #
    # Test date helpers
    #
    # --------------------------------------------------------------------------
    def test_today_yesterday_and_tomorrow(self):
        now = pysat.datetime.now()
        today = pysat.datetime(now.year, now.month, now.day)
        assert today == self.testInst.today()
        assert today - pds.DateOffset(days=1) == self.testInst.yesterday()
        assert today + pds.DateOffset(days=1) == self.testInst.tomorrow()

    def test_filter_datetime(self):
        now = pysat.datetime.now()
        today = pysat.datetime(now.year, now.month, now.day)
        assert today == filter_datetime_input(now)

    def test_filtered_date_attribute(self):
        now = pysat.datetime.now()
        today = pysat.datetime(now.year, now.month, now.day)
        self.testInst.date = now
        assert today == self.testInst.date

    # -------------------------------------------------------------------------
    #
    # Test concat_data method
    #
    # --------------------------------------------------------------------------

    def test_concat_data(self):
        # data set #2
        self.testInst.load(2009, 2)
        data2 = self.testInst.data
        len2 = len(self.testInst.index)
        self.testInst.load(2009, 1)
        # data set #1
        data1 = self.testInst.data
        len1 = len(self.testInst.index)

        # concat together
        self.testInst.data = self.testInst.concat_data([data1, data2])
        # basic test for concatenation
        len3 = len(self.testInst.index)
        assert (len3 == len1 + len2)

        if self.testInst.pandas_format:
            # test concat from above
            assert (self.testInst[0:len1, :] == data1.values[:, :]).all().all()
            assert (self.testInst[len1:, :] == data2.values[:, :]).all().all()
            # concat together with sort=True
            # pandas only feature
            self.testInst.data = self.testInst.concat_data([data1, data2],
                                                           sort=True)
            # test for concatenation
            len3 = len(self.testInst.index)
            assert (len3 == len1 + len2)
            assert (self.testInst[0:len1, data1.columns] == data1.values[:, :]).all().all()
            assert (self.testInst[len1:, data2.columns] == data2.values[:, :]).all().all()
        else:

            # first, check for concat just before if else
            assert (self.testInst[0:len1, :] == data1.to_array()[:, :]).all().all()
            assert (self.testInst[len1:, :] == data2.to_array()[:, :]).all().all()

            # concat together while also specifying a different concatentation dimension
            # xarray specific functionality
            # change name of main dim to support test for dim keyword
            data1 = data1.rename({'time': 'time2'})
            data2 = data2.rename({'time': 'time2'})

            # concat together
            self.testInst.data = \
                self.testInst.concat_data([data1, data2],
                                          dim='time2').rename({'time2':
                                                               'time'})
            # test for concatenation
            # Instrument.data must have a 'time' index
            len3 = len(self.testInst.index)
            assert (len3 == len1 + len2)
            assert (self.testInst[0:len1, :] == data1.to_array()[:, :]).all().all()
            assert (self.testInst[len1:, :] == data2.to_array()[:, :]).all().all()

    # --------------------------------------------------------------------------
    #
    # Test empty property flags, if True, no data
    #
    # --------------------------------------------------------------------------
    def test_empty_flag_data_empty(self):
        assert self.testInst.empty

    def test_empty_flag_data_not_empty(self):
        self.testInst.load(2009, 1)
        assert not self.testInst.empty

    # --------------------------------------------------------------------------
    #
    # Test index attribute, should always be a datetime index
    #
    # --------------------------------------------------------------------------
    def test_index_attribute(self):
        # empty Instrument test
        assert isinstance(self.testInst.index, pds.Index)
        # now repeat the same test but with data loaded
        self.testInst.load(2009, 1)
        assert isinstance(self.testInst.index, pds.Index)

    def test_index_return(self):
        # load data
        self.testInst.load(2009, 1)
        # ensure we get the index back
        if self.testInst.pandas_format:
            assert np.all(self.testInst.index == self.testInst.data.index)
        else:
            assert np.all(self.testInst.index ==
                          self.testInst.data.indexes['time'])

    #--------------------------------------------------------------------------
    #
    # Test custom attributes
    #
    #--------------------------------------------------------------------------



    @raises(AttributeError)
    def test_retrieve_bad_attribute(self):
        self.testInst.bad_attr

    def test_base_attr(self):
        self.testInst._base_attr
        assert '_base_attr' in dir(self.testInst)

    # --------------------------------------------------------------------------
    #
    # test textual representations
    #
    # --------------------------------------------------------------------------
    def test_basic_repr(self):
        """Check for lines from each decision point in repr"""
        output = self.testInst.__str__()
        assert isinstance(output, str)
        assert output.find('pysat Instrument object') > 0
        # No custom functions
        assert output.find('No functions applied') > 0
        # No orbital info
        assert output.find('Orbit properties not set') > 0
        # Files exist for test inst
        assert output.find('Date Range:') > 0
        # No loaded data
        assert output.find('No loaded data') > 0
        assert output.find('Number of variables:') < 0
        assert output.find('dummy') < 0

    def test_repr_w_orbit(self):
        re_load(pysat.instruments.pysat_testing)
        orbit_info = {'index': 'mlt',
                      'kind': 'local time',
                      'period': np.timedelta64(97, 'm')}
        testInst = pysat.Instrument(platform='pysat', name='testing',
                                    sat_id='10',
                                    clean_level='clean',
                                    update_files=True,
                                    orbit_info=orbit_info)

        output = testInst.__str__()
        # Check that orbit info is passed through
        assert output.find('Orbit properties not set') < 0
        assert output.find('Orbit Kind:') > 0
        assert output.find('Loaded Orbit Number: None') > 0
        # Activate orbits, check that message has changed
        testInst.load(2009, 1)
        testInst.orbits.next()
        output = testInst.__str__()
        assert output.find('Loaded Orbit Number: None') < 0
        assert output.find('Loaded Orbit Number: ') > 0

    def test_repr_w_padding(self):
        self.testInst.pad = pds.DateOffset(minutes=5)
        output = self.testInst.__str__()
        assert output.find('DateOffset: minutes=5') > 0

    def test_repr_w_custom_func(self):
        def testfunc(self):
            pass
        self.testInst.custom.add(testfunc, 'modify')
        output = self.testInst.__str__()
        assert output.find('testfunc') > 0

    def test_repr_w_load_data(self):
        self.testInst.load(2009, 1)
        output = self.testInst.__str__()
        assert output.find('No loaded data') < 0
        assert output.find('Number of variables:') > 0
        assert output.find('dummy') > 0

    # --------------------------------------------------------------------------
    #
    # test instrument initialization functions
    #
    # --------------------------------------------------------------------------
    def test_instrument_init(self):
        """Test if init function supplied by instrument can modify object"""
        assert self.testInst.new_thing

    def test_custom_instrument_load(self):
        """
        Test if the correct day is being loaded (End-to-End),
        with no instrument file but routines are passed.
        """
        import pysat.instruments.pysat_testing as test
        testInst = pysat.Instrument(inst_module=test, tag='',
                                    clean_level='clean')
        testInst.load(2009, 32)
        assert testInst.date == pds.datetime(2009, 2, 1)

    @raises(AttributeError)
    def test_custom_instrument_load_2(self):
        """
        Test if an exception is thrown correctly if there is no
        instrument file and supplied routines are incomplete.
        """
        import pysat.instruments.pysat_testing as test
        del test.list_files
        testIn = pysat.Instrument(inst_module=test, tag='',
                                  clean_level='clean')
        testIn.load(2009, 1)

    @raises(AttributeError)
    def test_custom_instrument_load_3(self):
        """
        Test if an exception is thrown correctly if there is no
        instrument file and supplied routines are incomplete.
        """
        import pysat.instruments.pysat_testing as test
        del test.load
        testIn = pysat.Instrument(inst_module=test, tag='',
                                  clean_level='clean')
        testIn.load(2009, 1)

    # --------------------------------------------------------------------------
    #
    # Test basis data access features, both getting and setting data
    #
    # --------------------------------------------------------------------------
    def test_basic_data_access_by_name(self):
        self.testInst.load(2009, 1)
        assert np.all(self.testInst['uts'] == self.testInst.data['uts'])

    def test_data_access_by_row_slicing_and_name(self):
        self.testInst.load(2009, 1)
        assert np.all(self.testInst[0:10, 'uts'] ==
                      self.testInst.data['uts'].values[0:10])

    def test_data_access_by_row_slicing_and_name_slicing(self):
        self.testInst.load(2009, 1)
        result = self.testInst[0:10, :]
        for variable, array in result.items():
            assert len(array) == len(self.testInst.data[variable].values[0:10])
            assert np.all(array == self.testInst.data[variable].values[0:10])

    def test_data_access_by_row_slicing_w_ndarray_and_name(self):
        self.testInst.load(2009, 1)
        assert np.all(self.testInst[np.arange(0, 10), 'uts'] ==
                      self.testInst.data['uts'].values[0:10])

    def test_data_access_by_row_and_name(self):
        self.testInst.load(2009, 1)
        assert np.all(self.testInst[0, 'uts'] ==
                      self.testInst.data['uts'].values[0])

    def test_data_access_by_row_index(self):
        self.testInst.load(2009, 1)
        idx = np.arange(10)
        assert np.all(self.testInst[idx]['uts'] ==
                      self.testInst.data['uts'].values[idx])

    def test_data_access_by_datetime_and_name(self):
        self.testInst.load(2009, 1)
        ind = pysat.datetime(2009, 1, 1, 0, 0, 0)
        assert np.all(self.testInst[ind, 'uts'] ==
                      self.testInst.data['uts'].values[0])

    def test_data_access_by_datetime_slicing_and_name(self):
        self.testInst.load(2009, 1)
        start = pysat.datetime(2009, 1, 1, 0, 0, 0)
        stop = pysat.datetime(2009, 1, 1, 0, 0, 10)
        assert np.all(self.testInst[start:stop, 'uts'] ==
                      self.testInst.data['uts'].values[0:11])

    def test_setting_data_by_name(self):
        self.testInst.load(2009, 1)
        self.testInst['doubleMLT'] = 2. * self.testInst['mlt']
        assert np.all(self.testInst['doubleMLT'] == 2. * self.testInst['mlt'])

    def test_setting_series_data_by_name(self):
        self.testInst.load(2009, 1)
        self.testInst['doubleMLT'] = 2.*pds.Series(self.testInst['mlt'].values,
                                                   index=self.testInst.index)
        assert np.all(self.testInst['doubleMLT'] == 2.*self.testInst['mlt'])

        self.testInst['blankMLT'] = pds.Series(None)
        assert np.all(np.isnan(self.testInst['blankMLT']))

    def test_setting_pandas_dataframe_by_names(self):
        self.testInst.load(2009, 1)
        self.testInst[['doubleMLT', 'tripleMLT']] = \
            pds.DataFrame({'doubleMLT': 2.*self.testInst['mlt'].values,
                           'tripleMLT': 3.*self.testInst['mlt'].values},
                          index=self.testInst.index)
        assert np.all(self.testInst['doubleMLT'] == 2.*self.testInst['mlt'])
        assert np.all(self.testInst['tripleMLT'] == 3.*self.testInst['mlt'])

    def test_setting_data_by_name_single_element(self):
        self.testInst.load(2009, 1)
        self.testInst['doubleMLT'] = 2.
        assert np.all(self.testInst['doubleMLT'] == 2.)

        self.testInst['nanMLT'] = np.nan
        assert np.all(np.isnan(self.testInst['nanMLT']))

    def test_setting_data_by_name_with_meta(self):
        self.testInst.load(2009, 1)
        self.testInst['doubleMLT'] = {'data': 2.*self.testInst['mlt'],
                                      'units': 'hours',
                                      'long_name': 'double trouble'}
        assert np.all(self.testInst['doubleMLT'] == 2.*self.testInst['mlt'])
        assert self.testInst.meta['doubleMLT'].units == 'hours'
        assert self.testInst.meta['doubleMLT'].long_name == 'double trouble'

    def test_setting_partial_data(self):
        self.testInst.load(2009, 1)
        cloneInst = self.testInst
        if self.testInst.pandas_format:
            self.testInst[0:3] = 0
            assert (np.all(self.testInst[3:] == cloneInst[3:]) &
                    np.all(self.testInst[0:3] == 0))
        else:
            # This command does not work for xarray
            assert True

    def test_setting_partial_data_by_name(self):
        self.testInst.load(2009, 1)
        self.testInst['doubleMLT'] = 2. * self.testInst['mlt']
        self.testInst[0, 'doubleMLT'] = 0
        assert np.all(self.testInst[1:, 'doubleMLT'] ==
                      2. * self.testInst[1:, 'mlt'])
        assert self.testInst[0, 'doubleMLT'] == 0

    def test_setting_partial_data_by_integer_and_name(self):
        self.testInst.load(2009, 1)
        self.testInst['doubleMLT'] = 2.*self.testInst['mlt']
        self.testInst[[0, 1, 2, 3], 'doubleMLT'] = 0
        assert np.all(self.testInst[4:, 'doubleMLT'] ==
                      2. * self.testInst[4:, 'mlt'])
        assert np.all(self.testInst[[0, 1, 2, 3], 'doubleMLT'] == 0)

    def test_setting_partial_data_by_slice_and_name(self):
        self.testInst.load(2009, 1)
        self.testInst['doubleMLT'] = 2. * self.testInst['mlt']
        self.testInst[0:10, 'doubleMLT'] = 0
        assert np.all(self.testInst[10:, 'doubleMLT'] ==
                      2. * self.testInst[10:, 'mlt'])
        assert np.all(self.testInst[0:10, 'doubleMLT'] == 0)

    def test_setting_partial_data_by_index_and_name(self):
        self.testInst.load(2009, 1)
        self.testInst['doubleMLT'] = 2. * self.testInst['mlt']
        self.testInst[self.testInst.index[0:10], 'doubleMLT'] = 0
        assert np.all(self.testInst[10:, 'doubleMLT'] ==
                      2. * self.testInst[10:, 'mlt'])
        assert np.all(self.testInst[0:10, 'doubleMLT'] == 0)

    def test_setting_partial_data_by_numpy_array_and_name(self):
        self.testInst.load(2009, 1)
        self.testInst['doubleMLT'] = 2. * self.testInst['mlt']
        self.testInst[np.array([0, 1, 2, 3]), 'doubleMLT'] = 0
        assert np.all(self.testInst[4:, 'doubleMLT'] ==
                      2. * self.testInst[4:, 'mlt'])
        assert np.all(self.testInst[0:4, 'doubleMLT'] == 0)

    def test_setting_partial_data_by_datetime_and_name(self):
        self.testInst.load(2009, 1)
        self.testInst['doubleMLT'] = 2. * self.testInst['mlt']
        self.testInst[pysat.datetime(2009, 1, 1, 0, 0, 0), 'doubleMLT'] = 0
        assert np.all(self.testInst[0, 'doubleMLT'] ==
                      2. * self.testInst[0, 'mlt'])
        assert np.all(self.testInst[0, 'doubleMLT'] == 0)

    def test_setting_partial_data_by_datetime_slicing_and_name(self):
        self.testInst.load(2009, 1)
        self.testInst['doubleMLT'] = 2. * self.testInst['mlt']
        self.testInst[pysat.datetime(2009, 1, 1, 0, 0, 0):
                      pysat.datetime(2009, 1, 1, 0, 0, 10),
                      'doubleMLT'] = 0
        assert np.all(self.testInst[11:, 'doubleMLT'] ==
                      2. * self.testInst[11:, 'mlt'])
        assert np.all(self.testInst[0:11, 'doubleMLT'] == 0)

    def test_modifying_data_inplace(self):
        self.testInst.load(2009, 1)
        self.testInst['doubleMLT'] = 2. * self.testInst['mlt']
        self.testInst['doubleMLT'] += 100
        assert np.all(self.testInst['doubleMLT'] ==
                      2.*self.testInst['mlt'] + 100)

    def test_getting_all_data_by_index(self):
        self.testInst.load(2009, 1)
        a = self.testInst[[0, 1, 2, 3, 4]]
        if self.testInst.pandas_format:
            assert len(a) == 5
        else:
            assert a.sizes['time'] == 5

    def test_getting_all_data_by_numpy_array_of_int(self):
        self.testInst.load(2009, 1)
        a = self.testInst[np.array([0, 1, 2, 3, 4])]
        if self.testInst.pandas_format:
            assert len(a) == 5
        else:
            assert a.sizes['time'] == 5

    # --------------------------------------------------------------------------
    #
    # Test iteration behaviors
    #
    # --------------------------------------------------------------------------
    @raises(StopIteration)
    def test_left_bounds_with_prev(self):
        """Test if passing bounds raises StopIteration."""
        # load first data
        self.testInst.next()
        # go back to no data
        self.testInst.prev()
        # self.testInst.prev()

    @raises(StopIteration)
    def test_right_bounds_with_next(self):
        """Test if passing bounds raises StopIteration."""
        # load last data
        self.testInst.prev()
        # move on to future data that doesn't exist
        self.testInst.next()

    def test_set_bounds_with_frequency(self):
        start = pysat.datetime(2009, 1, 1)
        stop = pysat.datetime(2010, 1, 15)
        self.testInst.bounds = (start, stop, 'M')
        assert np.all(self.testInst._iter_list == pds.date_range(start, stop,
                                                                 freq='M').tolist())

    @raises(Exception)
    def test_set_bounds_too_few(self):
        start = pysat.datetime(2009, 1, 1)
        self.testInst.bounds = [start]

    @raises(Exception)
    def test_set_bounds_mixed(self):
        start = pysat.datetime(2009, 1, 1)
        self.testInst.bounds = [start, '2009-01-01.nofile']

    @raises(Exception)
    def test_set_bounds_wrong_type(self):
        start = pysat.datetime(2009, 1, 1)
        self.testInst.bounds = [start, 1]

    @raises(Exception)
    def test_set_bounds_mixed_iterable(self):
        start = [pysat.datetime(2009, 1, 1)]*2
        self.testInst.bounds = [start, '2009-01-01.nofile']

    @raises(Exception)
    def test_set_bounds_mixed_iterabless(self):
        start = [pysat.datetime(2009, 1, 1)]*2
        self.testInst.bounds = [start, [pysat.datetime(2009, 1, 1), '2009-01-01.nofile']]

    def test_set_bounds_string_default_start(self):
        self.testInst.bounds = [None, '2009-01-01.nofile']
        assert self.testInst.bounds[0][0] == self.testInst.files[0]

    def test_set_bounds_string_default_end(self):
        self.testInst.bounds = ['2009-01-01.nofile', None]
        assert self.testInst.bounds[1][0] == self.testInst.files[-1]

    @raises(ValueError)
    def test_set_bounds_too_many(self):
        start = pysat.datetime(2009, 1, 1)
        stop = pysat.datetime(2009, 1, 1)
        huh = pysat.datetime(2009, 1, 1)
        self.testInst.bounds = [start, stop, huh]

    def test_set_bounds_by_date(self):
        start = pysat.datetime(2009, 1, 1)
        stop = pysat.datetime(2009, 1, 15)
        self.testInst.bounds = (start, stop)
        assert np.all(self.testInst._iter_list ==
                      pds.date_range(start, stop).tolist())

    def test_set_bounds_by_default(self):
        start = self.testInst.files.start_date
        stop = self.testInst.files.stop_date
        self.testInst.bounds = (None, None)
        assert np.all(self.testInst._iter_list ==
                      pds.date_range(start, stop).tolist())
        self.testInst.bounds = None
        assert np.all(self.testInst._iter_list ==
                      pds.date_range(start, stop).tolist())
        self.testInst.bounds = (start, None)
        assert np.all(self.testInst._iter_list ==
                      pds.date_range(start, stop).tolist())
        self.testInst.bounds = (None, stop)
        assert np.all(self.testInst._iter_list ==
                      pds.date_range(start, stop).tolist())

    def test_set_bounds_by_date_extra_time(self):
        start = pysat.datetime(2009, 1, 1, 1, 10)
        stop = pysat.datetime(2009, 1, 15, 1, 10)
        self.testInst.bounds = (start, stop)
        start = filter_datetime_input(start)
        stop = filter_datetime_input(stop)
        assert np.all(self.testInst._iter_list ==
                      pds.date_range(start, stop).tolist())

    def test_iterate_over_bounds_set_by_date(self):
        start = pysat.datetime(2009, 1, 1)
        stop = pysat.datetime(2009, 1, 15)
        self.testInst.bounds = (start, stop)
        dates = []
        for inst in self.testInst:
            dates.append(inst.date)
        out = pds.date_range(start, stop).tolist()
        assert np.all(dates == out)

    def test_iterate_over_bounds_set_by_date2(self):
        start = pysat.datetime(2008, 1, 1)
        stop = pysat.datetime(2010, 12, 31)
        self.testInst.bounds = (start, stop)
        dates = []
        for inst in self.testInst:
            dates.append(inst.date)
        out = pds.date_range(start, stop).tolist()
        assert np.all(dates == out)

    def test_iterate_over_default_bounds(self):
        start = self.testInst.files.start_date
        stop = self.testInst.files.stop_date
        self.testInst.bounds = (None, None)
        dates = []
        for inst in self.testInst:
            dates.append(inst.date)
        out = pds.date_range(start, stop).tolist()
        assert np.all(dates == out)

    def test_set_bounds_by_date_season(self):
        start = [pysat.datetime(2009, 1, 1), pysat.datetime(2009, 2, 1)]
        stop = [pysat.datetime(2009, 1, 15), pysat.datetime(2009, 2, 15)]
        self.testInst.bounds = (start, stop)
        out = pds.date_range(start[0], stop[0]).tolist()
        out.extend(pds.date_range(start[1], stop[1]).tolist())
        assert np.all(self.testInst._iter_list == out)

    def test_set_bounds_by_date_season_extra_time(self):
        start = [pysat.datetime(2009, 1, 1, 1, 10),
                 pysat.datetime(2009, 2, 1, 1, 10)]
        stop = [pysat.datetime(2009, 1, 15, 1, 10),
                pysat.datetime(2009, 2, 15, 1, 10)]
        self.testInst.bounds = (start, stop)
        start = filter_datetime_input(start)
        stop = filter_datetime_input(stop)
        out = pds.date_range(start[0], stop[0]).tolist()
        out.extend(pds.date_range(start[1], stop[1]).tolist())
        assert np.all(self.testInst._iter_list == out)

    def test_iterate_over_bounds_set_by_date_season(self):
        start = [pysat.datetime(2009, 1, 1), pysat.datetime(2009, 2, 1)]
        stop = [pysat.datetime(2009, 1, 15), pysat.datetime(2009, 2, 15)]
        self.testInst.bounds = (start, stop)
        dates = []
        for inst in self.testInst:
            dates.append(inst.date)
        out = pds.date_range(start[0], stop[0]).tolist()
        out.extend(pds.date_range(start[1], stop[1]).tolist())
        assert np.all(dates == out)

    def test_iterate_over_bounds_set_by_date_season_extra_time(self):
        start = [pysat.datetime(2009, 1, 1, 1, 10),
                 pysat.datetime(2009, 2, 1, 1, 10)]
        stop = [pysat.datetime(2009, 1, 15, 1, 10),
                pysat.datetime(2009, 2, 15, 1, 10)]
        self.testInst.bounds = (start, stop)
        # filter
        start = filter_datetime_input(start)
        stop = filter_datetime_input(stop)
        # iterate
        dates = []
        for inst in self.testInst:
            dates.append(inst.date)
        out = pds.date_range(start[0], stop[0]).tolist()
        out.extend(pds.date_range(start[1], stop[1]).tolist())
        assert np.all(dates == out)

    def test_set_bounds_by_fname(self):
        start = '2009-01-01.nofile'
        stop = '2009-01-03.nofile'
        self.testInst.bounds = (start, stop)
        assert np.all(self.testInst._iter_list ==
                      ['2009-01-01.nofile', '2009-01-02.nofile',
                       '2009-01-03.nofile'])

    def test_iterate_over_bounds_set_by_fname(self):
        start = '2009-01-01.nofile'
        stop = '2009-01-15.nofile'
        start_d = pysat.datetime(2009, 1, 1)
        stop_d = pysat.datetime(2009, 1, 15)
        self.testInst.bounds = (start, stop)
        dates = []
        for inst in self.testInst:
            dates.append(inst.date)
        out = pds.date_range(start_d, stop_d).tolist()
        assert np.all(dates == out)

    def test_iterate_over_bounds_set_by_fname_via_next(self):
        start = '2009-01-01.nofile'
        stop = '2009-01-15.nofile'
        start_d = pysat.datetime(2009, 1, 1)
        stop_d = pysat.datetime(2009, 1, 15)
        self.testInst.bounds = (start, stop)
        dates = []
        loop_next = True
        while loop_next:
            try:
                self.testInst.next()
                dates.append(self.testInst.date)
            except StopIteration:
                loop_next = False
        out = pds.date_range(start_d, stop_d).tolist()
        assert np.all(dates == out)

    def test_iterate_over_bounds_set_by_fname_via_prev(self):
        start = '2009-01-01.nofile'
        stop = '2009-01-15.nofile'
        start_d = pysat.datetime(2009, 1, 1)
        stop_d = pysat.datetime(2009, 1, 15)
        self.testInst.bounds = (start, stop)
        dates = []
        loop = True
        while loop:
            try:
                self.testInst.prev()
                dates.append(self.testInst.date)
            except StopIteration:
                loop = False
        out = pds.date_range(start_d, stop_d).tolist()
        assert np.all(dates == out[::-1])

    def test_set_bounds_by_fname_season(self):
        start = ['2009-01-01.nofile', '2009-02-01.nofile']
        stop = ['2009-01-03.nofile', '2009-02-03.nofile']
        self.testInst.bounds = (start, stop)
        assert np.all(self.testInst._iter_list ==
                      ['2009-01-01.nofile', '2009-01-02.nofile',
                       '2009-01-03.nofile', '2009-02-01.nofile',
                       '2009-02-02.nofile', '2009-02-03.nofile'])

    def test_iterate_over_bounds_set_by_fname_season(self):
        start = ['2009-01-01.nofile', '2009-02-01.nofile']
        stop = ['2009-01-15.nofile', '2009-02-15.nofile']
        start_d = [pysat.datetime(2009, 1, 1), pysat.datetime(2009, 2, 1)]
        stop_d = [pysat.datetime(2009, 1, 15), pysat.datetime(2009, 2, 15)]
        self.testInst.bounds = (start, stop)
        dates = []
        for inst in self.testInst:
            dates.append(inst.date)
        out = pds.date_range(start_d[0], stop_d[0]).tolist()
        out.extend(pds.date_range(start_d[1], stop_d[1]).tolist())
        assert np.all(dates == out)

    def test_creating_empty_instrument_object(self):
        null = pysat.Instrument()

        assert isinstance(null, pysat.Instrument)

    @raises(ValueError)
    def test_incorrect_creation_empty_instrument_object(self):
        # both name and platform should be empty
        _ = pysat.Instrument(platform='cnofs')

    @raises(AttributeError)
    def test_supplying_instrument_module_requires_name_and_platform(self):
        class Dummy:
            pass
        Dummy.name = 'help'

        _ = pysat.Instrument(inst_module=Dummy)


# ------------------------------------------------------------------------------
#
# Repeat tests above with xarray data
#
# ------------------------------------------------------------------------------
class TestBasicsXarray(TestBasics):
    def setup(self):
        re_load(pysat.instruments.pysat_testing_xarray)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing_xarray',
                                         sat_id='10',
                                         clean_level='clean',
                                         update_files=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


# ------------------------------------------------------------------------------
#
# Repeat TestBasics above with shifted file dates
#
# ------------------------------------------------------------------------------

class TestBasicsShiftedFileDates(TestBasics):
    def setup(self):
        re_load(pysat.instruments.pysat_testing)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat', name='testing',
                                         sat_id='10',
                                         clean_level='clean',
                                         update_files=True,
                                         mangle_file_dates=True,
                                         strict_time_flag=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


# ------------------------------------------------------------------------------
#
# Test Instrument with a non-unique and non-monotonic index
#
# ------------------------------------------------------------------------------


class TestMalformedIndex():
    def setup(self):
        re_load(pysat.instruments.pysat_testing)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat', name='testing',
                                         sat_id='10',
                                         clean_level='clean',
                                         malformed_index=True,
                                         update_files=True,
                                         strict_time_flag=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst

    # --------------------------------------------------------------------------
    #
    # Test checks on time uniqueness and monotonicity
    #
    # --------------------------------------------------------------------------
    @raises(ValueError)
    def test_ensure_unique_index(self):
        self.testInst.load(2009, 1)


# ------------------------------------------------------------------------------
#
# Repeat tests above with xarray data
#
# ------------------------------------------------------------------------------

class TestMalformedIndexXarray(TestMalformedIndex):
    def setup(self):
        re_load(pysat.instruments.pysat_testing_xarray)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing_xarray',
                                         sat_id='10',
                                         clean_level='clean',
                                         malformed_index=True,
                                         update_files=True,
                                         strict_time_flag=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


# ------------------------------------------------------------------------------
#
# Test data padding, loading by file
#
# ------------------------------------------------------------------------------
class TestDataPaddingbyFile():
    def setup(self):
        re_load(pysat.instruments.pysat_testing)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat', name='testing',
                                         clean_level='clean',
                                         pad={'minutes': 5},
                                         update_files=True)
        self.testInst.bounds = ('2008-01-01.nofile', '2010-12-31.nofile')

        self.rawInst = pysat.Instrument(platform='pysat', name='testing',
                                        clean_level='clean',
                                        update_files=True)
        self.rawInst.bounds = self.testInst.bounds

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst
        del self.rawInst

    def test_fid_data_padding(self):
        self.testInst.load(fid=1, verifyPad=True)
        self.rawInst.load(fid=1)
        assert ((self.testInst.index[0] ==
                 self.rawInst.index[0] - pds.DateOffset(minutes=5)) &
                (self.testInst.index[-1] ==
                 self.rawInst.index[-1] + pds.DateOffset(minutes=5)))

    def test_fid_data_padding_next(self):
        self.testInst.load(fid=1, verifyPad=True)
        self.testInst.next(verifyPad=True)
        self.rawInst.load(fid=2)
        assert ((self.testInst.index[0] ==
                 self.rawInst.index[0] - pds.DateOffset(minutes=5)) &
                (self.testInst.index[-1] ==
                 self.rawInst.index[-1] + pds.DateOffset(minutes=5)))

    def test_fid_data_padding_multi_next(self):
        """This also tests that _prev_data and _next_data cacheing"""
        self.testInst.load(fid=1)
        self.testInst.next()
        self.testInst.next(verifyPad=True)
        self.rawInst.load(fid=3)
        assert ((self.testInst.index[0] ==
                 self.rawInst.index[0] - pds.DateOffset(minutes=5)) &
                (self.testInst.index[-1] ==
                 self.rawInst.index[-1] + pds.DateOffset(minutes=5)))

    def test_fid_data_padding_prev(self):
        self.testInst.load(fid=2, verifyPad=True)
        self.testInst.prev(verifyPad=True)
        # print(self.testInst.index)
        self.rawInst.load(fid=1)
        # print(self.rawInst.index)
        # print(self.testInst.index[0], self.rawInst.index[0] -
        #   pds.DateOffset(minutes=5), self.testInst.index[-1],
        #   self.rawInst.index[-1] + pds.DateOffset(minutes=5))
        assert ((self.testInst.index[0] ==
                 self.rawInst.index[0] - pds.DateOffset(minutes=5)) &
                (self.testInst.index[-1] ==
                 self.rawInst.index[-1] + pds.DateOffset(minutes=5)))

    def test_fid_data_padding_multi_prev(self):
        """This also tests that _prev_data and _next_data cacheing"""
        self.testInst.load(fid=10)
        self.testInst.prev()
        self.testInst.prev(verifyPad=True)
        self.rawInst.load(fid=8)
        assert ((self.testInst.index[0] ==
                 self.rawInst.index[0] - pds.DateOffset(minutes=5)) &
                (self.testInst.index[-1] ==
                 self.rawInst.index[-1] + pds.DateOffset(minutes=5)))

    def test_fid_data_padding_jump(self):
        self.testInst.load(fid=1, verifyPad=True)
        self.testInst.load(fid=10, verifyPad=True)
        self.rawInst.load(fid=10)
        assert ((self.testInst.index[0] ==
                 self.rawInst.index[0] - pds.DateOffset(minutes=5)) &
                (self.testInst.index[-1] ==
                 self.rawInst.index[-1] + pds.DateOffset(minutes=5)))

    def test_fid_data_padding_uniqueness(self):
        self.testInst.load(fid=1, verifyPad=True)
        assert (self.testInst.index.is_unique)

    def test_fid_data_padding_all_samples_present(self):
        self.testInst.load(fid=1, verifyPad=True)
        test_index = pds.date_range(self.testInst.index[0],
                                    self.testInst.index[-1], freq='S')
        assert (np.all(self.testInst.index == test_index))

    def test_fid_data_padding_removal(self):
        self.testInst.load(fid=1)
        self.rawInst.load(fid=1)
        assert self.testInst.index[0] == self.rawInst.index[0]
        assert self.testInst.index[-1] == self.rawInst.index[-1]
        assert len(self.rawInst.data) == len(self.testInst.data)


# ------------------------------------------------------------------------------
#
# Repeat tests above with xarray data
#
# ------------------------------------------------------------------------------

class TestDataPaddingbyFileXarray():
    def setup(self):
        re_load(pysat.instruments.pysat_testing_xarray)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing_xarray',
                                         clean_level='clean',
                                         pad={'minutes': 5},
                                         update_files=True)
        self.testInst.bounds = ('2008-01-01.nofile', '2010-12-31.nofile')

        self.rawInst = pysat.Instrument(platform='pysat',
                                        name='testing_xarray',
                                        clean_level='clean',
                                        update_files=True)
        self.rawInst.bounds = self.testInst.bounds

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst
        del self.rawInst


class TestOffsetRightFileDataPaddingBasics(TestDataPaddingbyFile):
    def setup(self):
        re_load(pysat.instruments.pysat_testing)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat', name='testing',
                                         clean_level='clean',
                                         update_files=True,
                                         sim_multi_file_right=True,
                                         pad={'minutes': 5})
        self.rawInst = pysat.Instrument(platform='pysat', name='testing',
                                        tag='',
                                        clean_level='clean',
                                        update_files=True,
                                        sim_multi_file_right=True)
        self.testInst.bounds = ('2008-01-01.nofile', '2010-12-31.nofile')
        self.rawInst.bounds = self.testInst.bounds


class TestOffsetRightFileDataPaddingBasicsXarray(TestDataPaddingbyFile):
    def setup(self):
        re_load(pysat.instruments.pysat_testing_xarray)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing_xarray',
                                         clean_level='clean',
                                         update_files=True,
                                         sim_multi_file_right=True,
                                         pad={'minutes': 5})
        self.rawInst = pysat.Instrument(platform='pysat',
                                        name='testing_xarray',
                                        clean_level='clean',
                                        update_files=True,
                                        sim_multi_file_right=True)
        self.testInst.bounds = ('2008-01-01.nofile', '2010-12-31.nofile')
        self.rawInst.bounds = self.testInst.bounds


class TestOffsetLeftFileDataPaddingBasics(TestDataPaddingbyFile):
    def setup(self):
        re_load(pysat.instruments.pysat_testing)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat', name='testing',
                                         clean_level='clean',
                                         update_files=True,
                                         sim_multi_file_left=True,
                                         pad={'minutes': 5})
        self.rawInst = pysat.Instrument(platform='pysat', name='testing',
                                        clean_level='clean',
                                        update_files=True,
                                        sim_multi_file_left=True)
        self.testInst.bounds = ('2008-01-01.nofile', '2010-12-31.nofile')
        self.rawInst.bounds = self.testInst.bounds


class TestDataPadding():
    def setup(self):
        re_load(pysat.instruments.pysat_testing)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat', name='testing',
                                         clean_level='clean',
                                         pad={'minutes': 5},
                                         update_files=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst

    def test_data_padding(self):
        self.testInst.load(2009, 2, verifyPad=True)
        assert ((self.testInst.index[0] ==
                 self.testInst.date - pds.DateOffset(minutes=5)) &
                (self.testInst.index[-1] == self.testInst.date +
                 pds.DateOffset(hours=23, minutes=59, seconds=59) +
                 pds.DateOffset(minutes=5)))

    def test_data_padding_offset_instantiation(self):
        testInst = pysat.Instrument(platform='pysat', name='testing',
                                    clean_level='clean',
                                    pad=pds.DateOffset(minutes=5),
                                    update_files=True)
        testInst.load(2009, 2, verifyPad=True)
        assert ((testInst.index[0] ==
                 testInst.date - pds.DateOffset(minutes=5)) &
                (testInst.index[-1] == testInst.date +
                 pds.DateOffset(hours=23, minutes=59, seconds=59) +
                 pds.DateOffset(minutes=5)))

    @raises(Exception)
    def test_data_padding_bad_instantiation(self):
        testInst = pysat.Instrument(platform='pysat', name='testing',
                                    clean_level='clean',
                                    pad=2,
                                    update_files=True)

    def test_yrdoy_data_padding_missing_days(self):
        self.testInst.load(2008, 1)
        # test load
        self.testInst.load(2008, 0)
        # reset buffer data
        self.testInst.load(2008, -5)
        # test load, prev day empty, current and next has data
        self.testInst.load(2008, 1)
        # reset
        self.testInst.load(2008, -4)
        # etc
        self.testInst.load(2008, 2)
        self.testInst.load(2008, -3)
        self.testInst.load(2008, 3)
        # switch to missing data on the right
        self.testInst.load(2010, 365)
        self.testInst.load(2010, 360)
        self.testInst.load(2010, 366)
        self.testInst.load(2010, 360)
        self.testInst.load(2010, 367)
        assert True

    def test_data_padding_next(self):
        self.testInst.load(2009, 2, verifyPad=True)
        self.testInst.next(verifyPad=True)
        assert ((self.testInst.index[0] == self.testInst.date -
                 pds.DateOffset(minutes=5)) &
                (self.testInst.index[-1] == self.testInst.date +
                 pds.DateOffset(hours=23, minutes=59, seconds=59) +
                 pds.DateOffset(minutes=5)))

    def test_data_padding_multi_next(self):
        """This also tests that _prev_data and _next_data cacheing"""
        self.testInst.load(2009, 2)
        self.testInst.next()
        self.testInst.next(verifyPad=True)
        assert ((self.testInst.index[0] == self.testInst.date -
                 pds.DateOffset(minutes=5)) &
                (self.testInst.index[-1] == self.testInst.date +
                 pds.DateOffset(hours=23, minutes=59, seconds=59) +
                 pds.DateOffset(minutes=5)))

    def test_data_padding_prev(self):
        self.testInst.load(2009, 2, verifyPad=True)
        self.testInst.prev(verifyPad=True)
        print(self.testInst.index)
        assert ((self.testInst.index[0] == self.testInst.date -
                 pds.DateOffset(minutes=5)) &
                (self.testInst.index[-1] == self.testInst.date +
                 pds.DateOffset(hours=23, minutes=59, seconds=59) +
                 pds.DateOffset(minutes=5)))

    def test_data_padding_multi_prev(self):
        """This also tests that _prev_data and _next_data cacheing"""
        self.testInst.load(2009, 10)
        self.testInst.prev()
        self.testInst.prev(verifyPad=True)
        assert ((self.testInst.index[0] == self.testInst.date -
                 pds.DateOffset(minutes=5)) &
                (self.testInst.index[-1] == self.testInst.date +
                 pds.DateOffset(hours=23, minutes=59, seconds=59) +
                 pds.DateOffset(minutes=5)))

    def test_data_padding_jump(self):
        self.testInst.load(2009, 2, verifyPad=True)
        self.testInst.load(2009, 11, verifyPad=True)
        assert ((self.testInst.index[0] ==
                 self.testInst.date - pds.DateOffset(minutes=5)) &
                (self.testInst.index[-1] ==
                 self.testInst.date
                 + pds.DateOffset(hours=23, minutes=59, seconds=59)
                 + pds.DateOffset(minutes=5)))

    def test_data_padding_uniqueness(self):
        self.testInst.load(2009, 1, verifyPad=True)
        assert (self.testInst.index.is_unique)

    def test_data_padding_all_samples_present(self):
        self.testInst.load(2009, 1, verifyPad=True)
        test_index = pds.date_range(self.testInst.index[0],
                                    self.testInst.index[-1], freq='S')
        assert (np.all(self.testInst.index == test_index))

    def test_data_padding_removal(self):
        self.testInst.load(2009, 1)
        # print(self.testInst.index)
        assert (self.testInst.index[0] == self.testInst.date) & \
               (self.testInst.index[-1] == self.testInst.date +
                pds.DateOffset(hour=23, minutes=59, seconds=59))


class TestDataPaddingXarray(TestDataPadding):
    def setup(self):
        re_load(pysat.instruments.pysat_testing_xarray)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing_xarray',
                                         clean_level='clean',
                                         pad={'minutes': 5},
                                         update_files=True)


class TestMultiFileRightDataPaddingBasics(TestDataPadding):
    def setup(self):
        re_load(pysat.instruments.pysat_testing)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat', name='testing',
                                         clean_level='clean',
                                         update_files=True,
                                         sim_multi_file_right=True,
                                         pad={'minutes': 5},
                                         multi_file_day=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestMultiFileRightDataPaddingBasicsXarray(TestDataPadding):
    def setup(self):
        re_load(pysat.instruments.pysat_testing_xarray)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing_xarray',
                                         clean_level='clean',
                                         update_files=True,
                                         sim_multi_file_right=True,
                                         pad={'minutes': 5},
                                         multi_file_day=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestMultiFileLeftDataPaddingBasics(TestDataPadding):
    def setup(self):
        re_load(pysat.instruments.pysat_testing)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing',
                                         clean_level='clean',
                                         update_files=True,
                                         sim_multi_file_left=True,
                                         pad={'minutes': 5},
                                         multi_file_day=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestMultiFileLeftDataPaddingBasicsXarray(TestDataPadding):
    def setup(self):
        re_load(pysat.instruments.pysat_testing_xarray)
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing_xarray',
                                         clean_level='clean',
                                         update_files=True,
                                         sim_multi_file_left=True,
                                         pad={'minutes': 5},
                                         multi_file_day=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst



"""
tests the pysat averaging code
"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import warnings
import pysat
from pysat.ssnl import plot


class TestBasics():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean')
        self.testInst.bounds = (pysat.datetime(2008, 1, 1),
                                pysat.datetime(2008, 1, 1))

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst
        plt.close()

    def test_scatterplot_w_ioff(self):
        """Check if scatterplot generates"""

        plt.ioff()
        figs = plot.scatterplot(self.testInst, 'longitude', 'latitude',
                                'slt', [0.0, 24.0])

        axes = figs[0].get_axes()
        assert len(figs) == 1
        assert len(axes) == 3
        assert not mpl.is_interactive()

    def test_scatterplot_w_ion(self):
        """Check if scatterplot generates and resets to interactive mode"""

        plt.ion()
        figs = plot.scatterplot(self.testInst, 'longitude', 'latitude',
                                'slt', [0.0, 24.0])

        axes = figs[0].get_axes()
        assert len(figs) == 1
        assert len(axes) == 3
        assert mpl.is_interactive()

    def test_scatterplot_w_limits(self):
        """Check if scatterplot generates with appropriate limits"""

        figs = plot.scatterplot(self.testInst, 'longitude', 'latitude',
                                'slt', [0.0, 24.0],
                                xlim=[0, 360], ylim=[-80, 80])

        axes = figs[0].get_axes()
        assert len(figs) == 1
        assert len(axes) == 3
        assert axes[0].get_xlim() == (0, 360)
        assert axes[1].get_xlim() == (0, 360)
        assert axes[0].get_ylim() == (-80, 80)
        assert axes[1].get_ylim() == (-80, 80)

    def test_multiple_scatterplots(self):
        """Check if multiple scatterplots generate"""
        figs = plot.scatterplot(self.testInst, 'longitude', 'latitude',
                                ['slt', 'mlt'], [0.0, 24.0])

        axes = figs[0].get_axes()
        axes2 = figs[1].get_axes()
        assert len(figs) == 2
        assert len(axes) == 3
        assert len(axes2) == 3


class TestDeprecation():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        warnings.simplefilter("always")

    def teardown(self):
        """Runs after every method to clean up previous testing."""

    def test_deprecation_warning_scatterplot(self):
        """Test if scatterplot in ssnl is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            try:
                plot.scatterplot(None, 'longitude', 'latitude', ['slt', 'mlt'],
                                 [0.0, 24.0])
            except TypeError:
                # Setting inst to None should produce a TypeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning



# -*- coding: utf-8 -*-
"""Supports the Vector Electric Field Instrument (VEFI)
onboard the Communication and Navigation Outage Forecasting
System (C/NOFS) satellite. Downloads data from the
NASA Coordinated Data Analysis Web (CDAWeb).

Description from CDAWeb:

The DC vector magnetometer on the CNOFS spacecraft is a three axis, fluxgate
sensor with active thermal control situated on a 0.6m boom.  This magnetometer
measures the Earth's magnetic field using 16 bit A/D converters at 1 sample per
sec with a range of .. 45,000 nT.  Its primary objective on the CNOFS
spacecraft is to enable an accurate V x B measurement along the spacecraft
trajectory.  In order to provide an in-flight calibration of the magnetic field
data, we compare the most recent POMME model (the POtsdam Magnetic Model of the
Earth, http://geomag.org/models/pomme5.html) with the actual magnetometer
measurements to help determine a set of calibration parameters for the gains,
offsets, and non-orthogonality matrix of the sensor axes.  The calibrated
magnetic field measurements are provided in the data file here. The VEFI
magnetic field data file currently contains the following variables:
B_north   Magnetic field in the north direction
B_up      Magnetic field in the up direction
B_west    Magnetic field in the west direction

The data is PRELIMINARY, and as such, is intended for BROWSE PURPOSES ONLY.

References
----------
A brief discussion of the C/NOFS mission and instruments can be found at
de La Beaujardière, O., et al. (2004), C/NOFS: A mission to forecast
scintillations, J. Atmos. Sol. Terr. Phys., 66, 1573–1591,
doi:10.1016/j.jastp.2004.07.030.

Properties
----------
platform
    'cnofs'
name
    'vefi'
tag
    Select measurement type, one of {'dc_b'}
sat_id
    None supported

Note
----
- tag = 'dc_b': 1 second DC magnetometer data

Warnings
--------
- The data is PRELIMINARY, and as such, is intended for BROWSE PURPOSES ONLY.
- Limited cleaning routine.
- Module not written by VEFI team.

"""

from __future__ import print_function
from __future__ import absolute_import
import functools
import numpy as np

import pysat
from pysat.instruments.methods import nasa_cdaweb as cdw
from pysat.instruments.methods import general as mm_gen

platform = 'cnofs'
name = 'vefi'
tags = {'dc_b': 'DC Magnetometer data - 1 second'}
sat_ids = {'': ['dc_b']}
_test_dates = {'': {'dc_b': pysat.datetime(2009, 1, 1)}}

# support list files routine
# use the default CDAWeb method
fname = 'cnofs_vefi_bfield_1sec_{year:04d}{month:02d}{day:02d}_v05.cdf'
supported_tags = {'': {'dc_b': fname}}
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)
# support load routine
# use the default CDAWeb method
load = cdw.load

# support download routine
# use the default CDAWeb method
basic_tag = {'dir': '/pub/data/cnofs/vefi/bfield_1sec',
             'remote_fname': '{year:4d}/' + fname,
             'local_fname': fname}
supported_tags = {'': {'dc_b': basic_tag}}
download = functools.partial(cdw.download, supported_tags)
# support listing files currently on CDAWeb
list_remote_files = functools.partial(cdw.list_remote_files,
                                      supported_tags=supported_tags)


def clean(inst):
    """Routine to return VEFI data cleaned to the specified level

    Parameters
    -----------
    inst : (pysat.Instrument)
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    Notes
    --------
    'dusty' or 'clean' removes data when interpolation flag is set to 1

    """

    if (inst.clean_level == 'dusty') | (inst.clean_level == 'clean'):
        idx, = np.where(inst['B_flag'] == 0)
        inst.data = inst[idx, :]

    return None



# -*- coding: utf-8 -*-
"""
pysat.utils - utilities for running pysat
=========================================

pysat.utils contains a number of functions used
throughout the pysat package.  This includes conversion
of formats, loading of files, and user-supplied info
for the pysat data directory structure.
"""

from pysat.utils import coords, files, time, registry, testing
from pysat.utils._core import scale_units, load_netcdf4
from pysat.utils._core import NetworkLock, generate_instrument_list
from pysat.utils._core import available_instruments
from pysat.utils._core import display_available_instruments



"""
tests the pysat utils.time area
"""
import datetime as dt
import numpy as np

import pytest

from pysat.utils import time as pytime


class TestGetYearDay():

    def test_getyrdoy_1(self):
        """Test the date to year, day of year code functionality"""

        date = dt.datetime(2009, 1, 1)
        yr, doy = pytime.getyrdoy(date)

        assert ((yr == 2009) & (doy == 1))

    def test_getyrdoy_leap_year(self):
        """Test the date to year, day of year code functionality (leap_year)"""

        date = dt.datetime(2008, 12, 31)
        yr, doy = pytime.getyrdoy(date)

        assert ((yr == 2008) & (doy == 366))


class TestParseDate():

    def test_parse_date_2_digit_year(self):
        """Test the ability to parse a str to produce a pandas datetime"""

        date = pytime.parse_date('14', '10', '31')

        assert date == dt.datetime(2014, 10, 31)

    def test_parse_date_2_digit_year_last_century(self):
        """Test the ability to parse a str to produce a pandas datetime
        pre-2000"""

        date = pytime.parse_date('94', '10', '31', century=1900)

        assert date == dt.datetime(1994, 10, 31)

    def test_parse_date_4_digit_year(self):
        """Test the ability to parse a str to produce a pandas datetime"""

        date = pytime.parse_date('1994', '10', '31')

        assert date == dt.datetime(1994, 10, 31)

    def test_parse_date_bad_input(self):
        """Test the ability to identify a non-physical date"""

        with pytest.raises(ValueError):
            _ = pytime.parse_date('194', '15', '31')


class TestCalcFreq():

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.year = np.ones(4) * 2001
        self.month = np.ones(4) * 1

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.year, self.month

    def test_calc_freq(self):
        """Test index frequency calculation"""

        tind = pytime.create_datetime_index(year=self.year,
                                            month=self.month,
                                            uts=np.arange(0.0, 4.0, 1.0))
        freq = pytime.calc_freq(tind)

        assert freq.find("1S") == 0

    def test_calc_freq_ns(self):
        """Test index frequency calculation with nanosecond output"""

        tind = pytime.create_datetime_index(year=self.year,
                                            month=self.month,
                                            uts=np.arange(0.0, 0.04, .01))
        freq = pytime.calc_freq(tind)

        assert freq.find("10000000N") == 0

    def test_calc_freq_len_fail(self):
        """Test index frequency calculation with empty list"""

        with pytest.raises(ValueError):
            pytime.calc_freq(list())

    def test_calc_freq_type_fail(self):
        """Test index frequency calculation with non-datetime list"""

        with pytest.raises(AttributeError):
            pytime.calc_freq([1, 2, 3, 4])


class TestCreateDateRange():

    def test_create_date_range(self):
        """Test ability to generate season list"""

        start = dt.datetime(2012, 2, 28)
        stop = dt.datetime(2012, 3, 1)
        season = pytime.create_date_range(start, stop, freq='D')

        assert season[0] == start
        assert season[-1] == stop
        assert len(season) == 3

    def test_create_date_range_w_gaps(self):
        """Test ability to generate season list"""

        start = [dt.datetime(2012, 2, 28), dt.datetime(2013, 2, 28)]
        stop = [dt.datetime(2012, 3, 1), dt.datetime(2013, 3, 1)]
        season = pytime.create_date_range(start, stop, freq='D')

        assert season[0] == start[0]
        assert season[-1] == stop[-1]
        assert len(season) == 5


class TestCreateDatetimeIndex():

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.year = 2012 * np.ones(4)
        self.month = 2 * np.ones(4)
        self.day = 28 * np.ones(4)
        self.uts = np.arange(0, 4)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.year, self.month, self.day, self.uts

    def test_create_datetime_index(self):
        """Tests ability to create an array of datetime objects from distinct
        arrays of input parameters"""

        dates = pytime.create_datetime_index(year=self.year, month=self.month,
                                             day=self.day, uts=self.uts)

        assert dates[0] == dt.datetime(2012, 2, 28)
        assert dates[-1] == dt.datetime(2012, 2, 28, 0, 0, 3)
        assert len(dates) == 4

    def test_create_datetime_index_wo_year(self):
        """Must include a year"""

        with pytest.raises(ValueError):
            _ = pytime.create_datetime_index()

    def test_create_datetime_index_wo_month_day_uts(self):
        """Tests ability to generate missing parameters"""

        dates = pytime.create_datetime_index(year=self.year)

        assert dates[0] == dt.datetime(2012, 1, 1)
        assert dates[-1] == dt.datetime(2012, 1, 1)
        assert len(dates) == 4
